/*
 * MUST (Marmot Umpire Scalable Tool)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2011-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2011-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file BlockingCompletion.cpp
 *       @see must::BlockingCompletion.
 *
 *  @date 09.08.2011
 *  @author Tobias Hilbrich, Mathias Korepkat, Joachim Protze
 */

#include "BlockingCompletion.h"
#include "BlockingState.h"

#include <sstream>

using namespace must;

//=============================
// Constructor -- RequestWaitInfo
//=============================
RequestWaitInfo::RequestWaitInfo()
 : isCompleted (false),
   request (0),
   info (NULL)
{
    //Nothing to do
}

//=============================
// Destructor -- RequestWaitInfo
//=============================
RequestWaitInfo::~RequestWaitInfo()
{
    if (info) delete info;
    info = NULL;
}

//=============================
// Constructor -- WfgInfo
//=============================
WfgInfo::WfgInfo ()
 : type (ARC_AND),
   isMixed (false),
   subNodeToReq ()
{
    //Nothing to do
}

//=============================
// Constructor
//=============================
BlockingCompletion::BlockingCompletion (
                BlockingState* state,
                MustParallelId pId,
                MustLocationId lId,
                MustRequestType request)
: BlockingOp (state, pId, lId),
  myRequest (),
  myRequests (),
  myIsForAll (true),
  myNumCompleted (0),
  myMatchIndex (-1),
  hadAnActualCompletion (false),
  myWfgInfo (NULL)
{
    I_Request* rInfo = state->myRTrack->getRequest(pId, request);

    myRequest.request = request;

    //We are already done if this is an unknown, NULL, or inactive request
    if (!rInfo || rInfo->isNull() || !rInfo->isActive() || rInfo->isProcNull())
    {
        myNumCompleted = 1;
        myRequest.isCompleted = true;
    }
}

//=============================
// Constructor
//=============================
BlockingCompletion::BlockingCompletion (
                BlockingState* state,
                MustParallelId pId,
                MustLocationId lId,
                int count,
                MustRequestType *requests,
                bool isForAll)
: BlockingOp (state, pId, lId),
  myRequest (),
  myRequests (),
  myIsForAll (isForAll),
  myNumCompleted (0),
  myMatchIndex (-1),
  hadAnActualCompletion (false),
  myWfgInfo (NULL)
{
    myRequests.resize(count);
    for (int i = 0; i < count;i++)
    {
        I_Request* rInfo = state->myRTrack->getRequest(pId, requests[i]);

        myRequests[i].request = requests[i];

        //We are already done if this is an unknown, NULL, or inactive request
        if (!rInfo || rInfo->isNull() || !rInfo->isActive() || rInfo->isProcNull())
        {
            myNumCompleted++;
            myRequests[i].isCompleted = true;
        }
    }

    //If count was 0 we will think this was a single request not an array, we prepare for that
    if (count == 0)
    {
        myNumCompleted = 1;
        myRequest.isCompleted = true;
    }
}

//=============================
// Destructor
//=============================
BlockingCompletion::~BlockingCompletion (void)
{
    myRequests.clear();

    if (myWfgInfo) delete myWfgInfo;
    myWfgInfo = NULL;
}


//=============================
// process
//=============================
PROCESSING_RETURN BlockingCompletion::process (int rank)
{
    myState->applyNewCompletionOp(this);
    return PROCESSING_SUCCESS;
}

//=============================
// print
//=============================
GTI_RETURN BlockingCompletion::print (std::ostream &out)
{
    out << "Completion with ";

    if (myRequests.size() == 0)
        out << "1";
    else
        out << myRequests.size();

    out << " requests of which " << myNumCompleted << " completed.";

    return GTI_SUCCESS;
}

//=============================
// offerMatchedSend
//=============================
bool BlockingCompletion::offerMatchedSend (bool hasRequest, MustRequestType request)
{
    if (!hasRequest)
        return false;

    return addMatchedRequest(request);
}

//=============================
// offerMatchedReceive
//=============================
bool BlockingCompletion::offerMatchedReceive (bool hasRequest, MustRequestType request)
{
    if (!hasRequest)
        return false;

    return addMatchedRequest(request);
}

//=============================
// offerMatchedCollective
//=============================
bool BlockingCompletion::offerMatchedCollective (void)
{
    //We are not interested in these
    return false;
}

//=============================
// canComplete
//=============================
bool BlockingCompletion::canComplete (void)
{
    if (myRequests.size() == 0)
        return myNumCompleted == 1;

    if (myIsForAll)
        return myNumCompleted == myRequests.size();

    return hadAnActualCompletion;
}

//=============================
// addMatchedRequest
//=============================
bool BlockingCompletion::addMatchedRequest (MustRequestType request)
{
    if (canComplete()) return false;

    //IMPORTANT: invalidate the WFG info if something might change.
    dropWfgInfo ();

    if (myRequests.size() == 0)
    {
        if (myNumCompleted == 0 && request == myRequest.request)
        {
            myNumCompleted++;
            hadAnActualCompletion = true;
            return true;
        }
    }
    else
    {
        for (int i = 0; i < myRequests.size(); i++)
        {
            if (myRequests[i].isCompleted)
                continue;

            if (myRequests[i].request == request)
            {
                myNumCompleted++;
                hadAnActualCompletion = true;
                myRequests[i].isCompleted = true;

                if (myMatchIndex < 0)
                    myMatchIndex = i;

                return true;
            }
        }
    }

    return false;
}

//=============================
// isMixedOp
//=============================
bool BlockingCompletion::isMixedOp (void)
{
    initWfgInfo ();
    if (canComplete()) return false;

    return myWfgInfo->isMixed;
}

//=============================
// getWaitType
//=============================
ArcType BlockingCompletion::getWaitType (void)
{
    initWfgInfo ();
    if (canComplete()) return ARC_AND;

    return myWfgInfo->type;
}

//=============================
// mixedOpGetNumSubNodes
//=============================
int BlockingCompletion::mixedOpGetNumSubNodes (void)
{
    initWfgInfo ();
    if (canComplete()) return 0;

    return myWfgInfo->subNodeToReq.size();
}

//=============================
// getWaitedForRanks
//=============================
std::list<int> BlockingCompletion::getWaitedForRanks (std::list<std::string> *outLabels, std::list<std::pair<bool, std::pair<MustParallelId,MustLocationId> > > *pReferences)
{
    std::list<int> ret;

    //Init
    initWfgInfo ();
    if (canComplete()) return ret;

    //Process
    if (myRequests.size() == 0)
    {
        applyP2PToWait ("request", myRequest.info, &ret, outLabels, pReferences);
        return ret;
    }//If just a single request

    //Array of requests
    for (int i = 0; i < myRequests.size(); i++)
    {
        if (myRequests[i].isCompleted)
            continue;

        //Is this a request of a sub node for a split op ?
        if (myWfgInfo->isMixed && myRequests[i].info->isWc)
            continue;

        std::stringstream stream;
        stream << "[" << i << "]";
        applyP2PToWait (stream.str(), myRequests[i].info, &ret, outLabels, pReferences);
    }

    return ret;
}

//=============================
// getSubNodeWaitedForRanks
//=============================
std::list<int> BlockingCompletion::getSubNodeWaitedForRanks (int subId, std::string *outLabel, bool *outHasReference, MustParallelId *outPId, MustLocationId *outLId)
{
    std::list<int> ret;

    initWfgInfo ();
    if (canComplete()) return ret;

    if (subId >= myWfgInfo->subNodeToReq.size()) return ret;

    int rIndex = myWfgInfo->subNodeToReq[subId];

    std::stringstream stream;
    stream << "[" << rIndex << "]";
    if (outLabel) *outLabel = stream.str();
    if (outHasReference) *outHasReference = true;
    if (outPId) *outPId = myRequests[rIndex].info->pId;
    if (outLId) *outLId = myRequests[rIndex].info->lId;

    applyP2PToWait ("", myRequests[rIndex].info, &ret, NULL, NULL);

    return ret;
}

//=============================
// initRequestInfo
//=============================
bool BlockingCompletion::initRequestInfo (MustRequestType request, P2PInfo **outInfo)
{
    if (!outInfo) return false; //invalid call
    if (*outInfo) return true;  //already initialized

    *outInfo = new P2PInfo ();
    if (!myState->myP2PMatch->getP2PInfo(myRank, request, *outInfo))
    {
        std::cerr << "Internal Error: BlockingState could not find information for a P2P op in P2PMatch!" << std::endl;
        assert (0);
    }

    return true;
}

//=============================
// initWfgInfo
//=============================
bool BlockingCompletion::initWfgInfo (void)
{
    if (myWfgInfo) return true; //Already initialized
    if (canComplete()) return true; //Already completed, we should not do this at all!

    myWfgInfo = new WfgInfo ();

    //A single request (no array)
    if (myRequests.size() == 0)
    {
        initRequestInfo (myRequest.request, &(myRequest.info));

        myWfgInfo->isMixed = false;
        myWfgInfo->type = ARC_AND;

        if (myRequest.info->isWc)
            myWfgInfo->type = ARC_OR;
        return true;
    }

    //Expected type
    myWfgInfo->type = ARC_AND;
    if (!myIsForAll)
        myWfgInfo->type = ARC_OR;;
    myWfgInfo->isMixed = false;

    //Array: Initialize all infos
    for (int i = 0; i < myRequests.size(); i++)
    {
        if (myRequests[i].isCompleted)
            continue;

        initRequestInfo (myRequests[i].request, &(myRequests[i].info));

        //Do we need to mix ?
        //Expected type
        if (myRequests.size() > 1 && myRequests[i].info->isWc && myWfgInfo->type == ARC_AND)
        {
            myWfgInfo->isMixed = true;
            myWfgInfo->subNodeToReq.insert (std::make_pair (myWfgInfo->subNodeToReq.size(), i));
        }

        //Special case a Waitall with exactly one wc receive request -> we change the expected type in this case to OR
        if (myRequests.size() == 1 && myRequests[i].info->isWc && myWfgInfo->type == ARC_AND)
        {
            myWfgInfo->type == ARC_OR;
        }
    }

    return true;
}

//=============================
// dropWfgInfo
//=============================
bool BlockingCompletion::dropWfgInfo (void)
{
    if (myWfgInfo) delete myWfgInfo;
    myWfgInfo = NULL;
    return true;
}

/*EOF*/
