/*
 * MUST (Marmot Umpire Scalable Tool)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2011-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2011-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file OperationReordering.cpp
 *       @see MUST::OperationReordering.
 *
 *  @date 25.07.2011
 *  @author Tobias Hilbrich, Mathias Korepkat, Joachim Protze
 */

#include "GtiMacros.h"

#include "OperationReordering.h"

using namespace must;

mGET_INSTANCE_FUNCTION(OperationReordering);
mFREE_INSTANCE_FUNCTION(OperationReordering);
mPNMPI_REGISTRATIONPOINT_FUNCTION(OperationReordering);

//=============================
// Constructor
//=============================
OperationReordering::OperationReordering (const char* instanceName)
    : gti::ModuleBase<OperationReordering, I_OperationReordering> (instanceName),
      myRankBlocked (),
      myIsSuspended (false),
      myQueues (),
      myOpenNonEmptyQueueIndices (),
      myRankinNonEmptyQueueList (),
      myIndexInOpenNonEmptyIndices (),
      myIsInProcessing (false),
      myNumOps (0)
{
    //Nothing to do
}

//=============================
// ~OperationReordering
//=============================
OperationReordering::~OperationReordering (void)
{
#ifdef MUST_DEBUG
    for (int i = 0; i < myQueues.size(); i++)
    {

        if (myQueues[i].size())
        {
            std::cout << "Warning: OperationReordering has " << myQueues[i].size() << " operation queued for rank " << i << " during its destruction:" << std::endl;

            std::list<I_Operation*>::iterator iter;
            int j = 1;
            while (!myQueues[i].empty())
            {
                I_Operation* op = myQueues[i].front();
                myQueues[i].pop();
                j++;

                std::cout << "- (" << j << "):";
                //We do not call print here as the operation may have inconsistent data due to the ongoing destruction
                //if (*iter) (*iter)->print (std::cout);
                std::cout << op << std::endl;
            }
        }
    }
#endif

    myQueues.clear ();
    myRankBlocked.clear();
    myRankinNonEmptyQueueList.clear ();
    myOpenNonEmptyQueueIndices.clear ();
    myIndexInOpenNonEmptyIndices.clear ();
}

//=============================
// init
//=============================
GTI_ANALYSIS_RETURN OperationReordering::init (
        int worldSize)
{
    if (!myQueues.size())
    {
        myQueues.resize(worldSize);
        myRankBlocked.resize(worldSize, false);
        myRankinNonEmptyQueueList.resize(worldSize, false);
        myIndexInOpenNonEmptyIndices.resize(worldSize);
    }
    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// isRankOpen
//=============================
bool OperationReordering::isRankOpen (int rank)
{
    if (rank >= myRankBlocked.size())
    {
        std::cerr << "ERROR: initialization failed, OperationReordering did not receive the init event." << std::endl;
        assert (0);
        return false;
    }

    if (myIsSuspended)
        return false;

    if (myRankBlocked[rank])
        return false;

    /*
     * Rank is also not open if there is at least one op
     * that is still queued for it!
     */
    return myQueues[rank].empty();
}

//=============================
// blockRank
//=============================
GTI_RETURN OperationReordering::blockRank (int rank)
{
    if (rank >= myRankBlocked.size())
    {
        std::cerr << "ERROR: initialization failed, OperationReordering did not receive the init event." << std::endl;
        assert (0);
        return GTI_ERROR;
    }

    //Remove from list of open ranks if it was there
    if (myRankinNonEmptyQueueList[rank])
    {
        std::list<int>::iterator iter = myIndexInOpenNonEmptyIndices[rank];
        myOpenNonEmptyQueueIndices.erase(iter);
        myRankinNonEmptyQueueList[rank] = false;
    }

    myRankBlocked[rank] = true;
    return GTI_SUCCESS;
}

//=============================
// resumeRank
//=============================
GTI_RETURN OperationReordering::resumeRank (int rank)
{
    if (rank >= myRankBlocked.size())
    {
        std::cerr << "ERROR: initialization failed, OperationReordering did not receive the init event." << std::endl;
        assert (0);
        return GTI_ERROR;
    }

    myRankBlocked[rank] = false;

    //Add to open ranks if it has queued ops and is not already there
    if (!myQueues[rank].empty() && !myRankinNonEmptyQueueList[rank])
    {
        myRankinNonEmptyQueueList[rank] = true;
        myOpenNonEmptyQueueIndices.push_front(rank);
        myIndexInOpenNonEmptyIndices[rank] = myOpenNonEmptyQueueIndices.begin();
    }

    return processQueues ();
}

//=============================
// enqueueOp
//=============================
GTI_RETURN OperationReordering::enqueueOp (int rank, I_Operation* op)
{
    if (rank >= myRankBlocked.size())
    {
        std::cerr << "ERROR: initialization failed, OperationReordering did not receive the init event." << std::endl;
        assert (0);
        return GTI_ERROR;
    }

    myQueues[rank].push(op);
    myNumOps++;

    //Do we need to add the rank into the non-empty queue list ?
    if (!myRankBlocked[rank] && !myRankinNonEmptyQueueList[rank])
    {
        myRankinNonEmptyQueueList[rank] = true;
        myOpenNonEmptyQueueIndices.push_front(rank);
        myIndexInOpenNonEmptyIndices[rank] = myOpenNonEmptyQueueIndices.begin();
    }

    return GTI_SUCCESS;
}

//=============================
// suspend
//=============================
GTI_RETURN OperationReordering::suspend (void)
{
    if (!myRankBlocked.size())
    {
        std::cerr << "ERROR: initialization failed, OperationReordering did not receive the init event." << std::endl;
        assert (0);
        return GTI_ERROR;
    }

    myIsSuspended = true;
    return GTI_SUCCESS;
}

//=============================
// removeSuspension
//=============================
GTI_RETURN OperationReordering::removeSuspension (void)
{
    if (!myRankBlocked.size())
    {
        std::cerr << "ERROR: initialization failed, OperationReordering did not receive the init event." << std::endl;
        assert (0);
        return GTI_ERROR;
    }

    myIsSuspended = false;
    return processQueues();
}

//=============================
// processQueues
//=============================
GTI_RETURN OperationReordering::processQueues (void)
{
    //Protection against recursive calling of processQueues, not thread safe!
    if (myIsInProcessing)
        return GTI_SUCCESS;

    myIsInProcessing = true;

    while (!myIsSuspended && !myOpenNonEmptyQueueIndices.empty())
    {
        int rank = myOpenNonEmptyQueueIndices.front();

        //Is there still something to process ?
        if (myRankBlocked[rank] || myQueues[rank].empty())
        {
            std::list<int>::iterator iter = myIndexInOpenNonEmptyIndices[rank];
            myOpenNonEmptyQueueIndices.erase(iter);
            myRankinNonEmptyQueueList[rank] = false;
            continue;
        }

        //Do the processing
        I_Operation* op = myQueues[rank].front();

        PROCESSING_RETURN ret = PROCESSING_SUCCESS;
        if (op)
            ret = op->process(rank);

        //Evaluate the processing return value
        if (ret == PROCESSING_ERROR) return GTI_ERROR;

        if (ret != PROCESSING_REEXECUTE)
        {
            myQueues[rank].pop();
            myNumOps--;

            if (myQueues[rank].empty() && myRankinNonEmptyQueueList[rank])
            {
                std::list<int>::iterator iter = myIndexInOpenNonEmptyIndices[rank];
                myOpenNonEmptyQueueIndices.erase(iter);
                myRankinNonEmptyQueueList[rank] = false;
            }
        }
    }

    myIsInProcessing = false;

    return GTI_SUCCESS;
}

//=============================
// getTotalQueueSize
//=============================
int OperationReordering::getTotalQueueSize (void)
{
    return myNumOps;
}

/*EOF*/
