/*
 * MUST (Marmot Umpire Scalable Tool)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2011-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2011-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file P2PMatch.cpp
 *       @see MUST::P2PMatch.
 *
 *  @date 26.07.2011
 *  @author Tobias Hilbrich, Joachim Protze, Mathias Korepkat
 */

#include "GtiMacros.h"
#include "MustEnums.h"

#include "P2PMatch.h"

#include <assert.h>
#include <sstream>
#include <fstream>

using namespace must;

mGET_INSTANCE_FUNCTION(P2PMatch);
mFREE_INSTANCE_FUNCTION(P2PMatch);
mPNMPI_REGISTRATIONPOINT_FUNCTION(P2PMatch);

//=============================
// Constructor
//=============================
P2PMatch::P2PMatch (const char* instanceName)
    : gti::ModuleBase<P2PMatch, I_P2PMatch> (instanceName),
      myQs (),
      mySuspendedForEntry (NULL),
      suspendedWcRecvs (),
      myFinCompletion (NULL),
      myListeners (),
      myPrintLostMessages (true)
#ifdef MUST_DEBUG
      ,myMatches ()
#endif
{
    //create sub modules
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //handle sub modules
#define NUM_MODS_REQUIRED 7
    if (subModInstances.size() < NUM_MODS_REQUIRED)
    {
    		std::cerr << "Module has not enough sub modules, check its analysis specification! (" << __FILE__ << "@" << __LINE__ << ")" << std::endl;
    		assert (0);
    }
    if (subModInstances.size() > NUM_MODS_REQUIRED)
    {
    		for (int i = NUM_MODS_REQUIRED; i < subModInstances.size(); i++)
    			destroySubModuleInstance (subModInstances[i]);
    }

    myPIdMod = (I_ParallelIdAnalysis*) subModInstances[0];
    myConsts = (I_BaseConstants*) subModInstances[1];
    myLogger = (I_CreateMessage*) subModInstances[2];
    myCTrack = (I_CommTrack*) subModInstances[3];
    myRTrack = (I_RequestTrack*) subModInstances[4];
    myDTrack = (I_DatatypeTrack*) subModInstances[5];
    myOrder = (I_OperationReordering*) subModInstances[6];

    //Initialize module data
    /*nothing to do*/
}

//=============================
// Destructor
//=============================
P2PMatch::~P2PMatch ()
{
	if (myPIdMod)
		destroySubModuleInstance ((I_Module*) myPIdMod);
	myPIdMod = NULL;

	if (myConsts)
	    destroySubModuleInstance ((I_Module*) myConsts);
	myConsts = NULL;

	if (myLogger)
		destroySubModuleInstance ((I_Module*) myLogger);
	myLogger = NULL;

	if (myCTrack)
		destroySubModuleInstance ((I_Module*) myCTrack);
	myCTrack = NULL;

	if (myRTrack)
		destroySubModuleInstance ((I_Module*) myRTrack);
	myRTrack = NULL;

	if (myDTrack)
	    destroySubModuleInstance ((I_Module*) myDTrack);
	myDTrack = NULL;

	if (myOrder)
	    destroySubModuleInstance ((I_Module*) myOrder);
	myOrder = NULL;

	//==Free other data
	//Matching structure
	//We do the graceful free in printLostMessages, we should not destruct anything here as destruction hell is already going on.
	myQs.clear ();

	//Completion
	if (myFinCompletion) delete (myFinCompletion);
}

//=============================
// send
//=============================
GTI_ANALYSIS_RETURN P2PMatch::send (
		MustParallelId pId,
		MustLocationId lId,
		int destIn,
		int tag,
		MustCommType comm,
		MustDatatypeType type,
		int count,
        int mode)
{
    //== 1) Prepare the operation for this send
    //MPI_PROC_NULL ops are no-ops!
	if (myConsts->isProcNull(destIn))
		return GTI_ANALYSIS_SUCCESS;

	//Get infos and translation
	int dest;
	I_CommPersistent* cInfo;
	I_DatatypePersistent *dInfo;
	if (!getCommTranslationAndType (pId, comm, destIn, type, &cInfo, &dest, &dInfo))
	    return GTI_ANALYSIS_SUCCESS;

	//Create op
	P2POp* newOp = new P2POp (
	        this,
	        true,
	        tag,
	        dest /*TRANSLATED dest*/,
	        cInfo,
	        dInfo,
	        count,
	        pId,
	        lId,
	        (MustSendMode) mode);

	//== 2) Process or Queue ?
	int rank = myPIdMod->getInfoForId(pId).rank;
	handleNewOp (rank, newOp);

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// isend
//=============================
GTI_ANALYSIS_RETURN P2PMatch::isend (
		MustParallelId pId,
		MustLocationId lId,
		int destIn,
		int tag,
		MustCommType comm,
		MustDatatypeType type,
		int count,
        int mode,
		MustRequestType request)
{
    //== 1) Prepare the operation for this send
    //MPI_PROC_NULL ops are no-ops!
    if (myConsts->isProcNull(destIn))
        return GTI_ANALYSIS_SUCCESS;

    //Get infos and translation
    int dest;
    I_CommPersistent* cInfo;
    I_DatatypePersistent *dInfo;
    if (!getCommTranslationAndType (pId, comm, destIn, type, &cInfo, &dest, &dInfo))
        return GTI_ANALYSIS_SUCCESS;

    //Create op
    P2POp* newOp = new P2POp (
            this,
            true,
            tag,
            dest /*TRANSLATED dest*/,
            request,
            cInfo,
            dInfo,
            count,
            pId,
            lId,
            (MustSendMode) mode);

    //== 2) Process or Queue ?
    int rank = myPIdMod->getInfoForId(pId).rank;
    handleNewOp (rank, newOp);

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// recv
//=============================
GTI_ANALYSIS_RETURN P2PMatch::recv (
		MustParallelId pId,
		MustLocationId lId,
		int sourceIn,
		int tag,
		MustCommType comm,
		MustDatatypeType type,
		int count)
{
    //== 1) Prepare the operation for this receive
    //MPI_PROC_NULL ops are no-ops!
    if (myConsts->isProcNull(sourceIn))
        return GTI_ANALYSIS_SUCCESS;

    //Get infos and translation
    int source;
    I_CommPersistent* cInfo;
    I_DatatypePersistent *dInfo;
    if (!getCommTranslationAndType (pId, comm, sourceIn, type, &cInfo, &source, &dInfo))
        return GTI_ANALYSIS_SUCCESS;

    //Create op
    P2POp* newOp = new P2POp (
            this,
            false,
            tag,
            source /*TRANSLATED source*/,
            cInfo,
            dInfo,
            count,
            pId,
            lId);

    //== 2) Process or Queue ?
    int rank = myPIdMod->getInfoForId(pId).rank;
    handleNewOp (rank, newOp);

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// irecv
//=============================
GTI_ANALYSIS_RETURN P2PMatch::irecv (
		MustParallelId pId,
		MustLocationId lId,
		int sourceIn,
		int tag,
		MustCommType comm,
		MustDatatypeType type,
		int count,
		MustRequestType request)
{
    //== 1) Prepare the operation for this receive
    //MPI_PROC_NULL ops are no-ops!
    if (myConsts->isProcNull(sourceIn))
        return GTI_ANALYSIS_SUCCESS;

    //Get infos and translation
    int source;
    I_CommPersistent* cInfo;
    I_DatatypePersistent *dInfo;
    if (!getCommTranslationAndType (pId, comm, sourceIn, type, &cInfo, &source, &dInfo))
        return GTI_ANALYSIS_SUCCESS;

    //Create op
    P2POp* newOp = new P2POp (
            this,
            false,
            tag,
            source /*TRANSLATED source*/,
            request,
            cInfo,
            dInfo,
            count,
            pId,
            lId);

    //== 2) Process or Queue ?
    int rank = myPIdMod->getInfoForId(pId).rank;
    handleNewOp (rank, newOp);

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// getCommTranslationAndType
//=============================
bool P2PMatch::getCommTranslationAndType (
        MustParallelId pId,
        MustCommType comm,
        int rankIn,
        MustDatatypeType type,
        I_CommPersistent**pOutComm,
        int *pOutTranslatedRank,
        I_DatatypePersistent**pOutType)
{
    //==1) Comm
    I_CommPersistent* cInfo = myCTrack->getPersistentComm(pId,comm);
    if (cInfo == NULL)
        return false;//Unknown comm
    if (cInfo->isNull())
    {
        cInfo->erase();
        return false;//MPI_COMM_NULL
    }

    if (pOutComm)
        *pOutComm = cInfo;

    //==2) Translation
    if (pOutTranslatedRank)
        *pOutTranslatedRank = translateDestSource (cInfo, rankIn);


    //==3) Datatype
    I_DatatypePersistent *dInfo = myDTrack->getPersistentDatatype (pId, type);
    if (!dInfo) return false; //Unknown datatype

    if (pOutType) *pOutType = dInfo;

    return true;
}

//=============================
// translateDestSource
//=============================
int P2PMatch::translateDestSource (I_Comm* comm, int destSource)
{
    int ret;
    if (destSource != myConsts->getAnySource())
    {
        if (!comm->isIntercomm())
            comm->getGroup()->translate(destSource, &ret);
        else
            comm->getRemoteGroup()->translate(destSource, &ret);
    }
    else
    {
        ret = destSource;
    }

    return ret;
}

//=============================
// invTranslateIssuingRank
//=============================
int P2PMatch::invTranslateIssuingRank (I_Comm* comm, int rank)
{
    int ret;
    if (rank != myConsts->getAnySource())
    {
        comm->getGroup()->containsWorldRank(rank, &ret);
    }
    else
    {
        ret = rank;
    }

    return ret;
}

//=============================
// invTranslateIssuingRank
//=============================
int P2PMatch::invTranslateDestSource (I_Comm* comm, int rank)
{
    int ret;
    if (rank != myConsts->getAnySource())
    {
        if (!comm->isIntercomm())
            comm->getGroup()->containsWorldRank(rank, &ret);
        else
            comm->getRemoteGroup()->containsWorldRank(rank, &ret);
    }
    else
    {
        ret = rank;
    }

    return ret;
}

//=============================
// handleNewOp
//=============================
void P2PMatch::handleNewOp (int rank, P2POp* op)
{
    if (myOrder->isRankOpen(rank))
    {
        //Process
        PROCESSING_RETURN ret = op->process(rank);

        //If processing caused a suspension and a re-processing we have to queue
        if (ret == PROCESSING_REEXECUTE)
        {
            myOrder->enqueueOp(rank, op);
            op->addToSuspendedWCOpQueue ();//add to queue of operations that are queued and have a wilcard if applicable to this op
        }
    }
    else
    {
        //Queue
        myOrder->enqueueOp(rank, op);
        op->addToSuspendedWCOpQueue ();//add to queue of operations that are queued and have a wilcard if applicable to this op
    }
}

//=============================
// recvUpdate
//=============================
GTI_ANALYSIS_RETURN P2PMatch::recvUpdate (
		MustParallelId pId,
		MustLocationId lId,
		int source)
{
	int from = myPIdMod->getInfoForId(pId).rank;

	//==Perform the update
	bool unsuspend = false;
	findRecvForUpdate (from, source, false, 0, &unsuspend);

#ifdef MUST_MATCH_DEBUG
	std::cout << "P2PMatch: WcRecvUpdate newSource=" << source << " from rank " << myPIdMod->getInfoForId(pId).rank << std::endl;
	printQs ();
#endif

	//Remove suspension if necessary
	if (unsuspend)
	    myOrder->removeSuspension();

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// irecvUpdate
//=============================
GTI_ANALYSIS_RETURN P2PMatch::irecvUpdate (
		MustParallelId pId,
		MustLocationId lId,
		int source,
		MustRequestType request)
{
	int from = myPIdMod->getInfoForId(pId).rank;

	//==Perform the update
	bool unsuspend = false;
	findRecvForUpdate (from, source, true, request, &unsuspend);

#ifdef MUST_MATCH_DEBUG
	std::cout << "LostMessage: WcIrecvUpdate newSource=" << source << " request=" << request << " from rank " << myPIdMod->getInfoForId(pId).rank << std::endl;
	printQs ();
#endif

	//Remove suspension if necessary
	if (unsuspend)
	    myOrder->removeSuspension();

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// startPersistent
//=============================
GTI_ANALYSIS_RETURN P2PMatch::startPersistent (
		MustParallelId pId,
		MustLocationId lId,
		MustRequestType request)
{
	I_Request* info = myRTrack->getRequest(pId, request);

	if (info == NULL || !info->isPersistent())
		return GTI_ANALYSIS_SUCCESS;

	/*We are mapped before RequestTrack, so we check here whether the request is still active!*/
	if (info->isActive())
	    return GTI_ANALYSIS_SUCCESS;

	int destSource;
	if (info->isSend())
	    destSource = info->getDest();
	else
	    destSource = info->getSource();

	//Don't do anything for ops with MPI_PROC_NULL
	if (destSource == myConsts->getProcNull())
	    return GTI_ANALYSIS_SUCCESS;

	I_CommPersistent* comm = info->getCommCopy();
	I_DatatypePersistent* type = info->getDatatypeCopy();

	int destSourceTranslate = translateDestSource (comm, destSource);

	P2POp* newOp = new P2POp (
            this,
            info->isSend(),
            info->getTag(),
            destSourceTranslate,
            request,
            comm,
            type,
            info->getCount(),
            pId,
            lId,
            info->getSendMode()
            );

	handleNewOp (newOp->getIssuerRank(), newOp);

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// cancel
//=============================
GTI_ANALYSIS_RETURN P2PMatch::cancel (
		MustParallelId pId,
		MustLocationId lId,
		MustRequestType request)
{
	static bool warned = false;
	if (!warned)
		std::cerr << "P2PMatch: detected a cancel, not supported, outputs may be wrong!" << std::endl;
	warned = true;

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// printLostMessages
//=============================
GTI_ANALYSIS_RETURN P2PMatch::printLostMessages (
		I_ChannelId *thisChannel)
{
	//Is completely reduced without a channel id ? (Reduced on this place)
	if (thisChannel)
	{
		//Initialize completion tree
		if (!myFinCompletion)
			myFinCompletion = new CompletionTree (
					thisChannel->getNumUsedSubIds()-1, thisChannel->getSubIdNumChannels(thisChannel->getNumUsedSubIds()-1));

		myFinCompletion->addCompletion (thisChannel);
	}

	if (!thisChannel || myFinCompletion->isCompleted())
	{
		//Log the lost messages
		printLostMessages ();
	}

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// printLostMessages
//=============================
GTI_ANALYSIS_RETURN P2PMatch::printLostMessages ()
{
	QT::iterator qIter;

	if (myPrintLostMessages)
	{
        for (qIter = myQs.begin(); qIter != myQs.end(); qIter++)
        {
            ProcessQueues::iterator cIter;
            for (cIter = qIter->second.begin(); cIter != qIter->second.end(); cIter++)
            {
                ProcessTable *t = &(cIter->second);
                RankOutstandings::iterator oIter;

                //==== SENDS
                for (oIter = t->sends.begin(); oIter != t->sends.end(); oIter++)
                {
                    std::list<P2POp*>::iterator lIter;
                    for (lIter = oIter->second.begin(); lIter != oIter->second.end(); lIter++)
                    {
                        if (*lIter) (*lIter)->logAsLost(qIter->first);
                    }
                }

                //==== RECVS
                for (oIter = t->recvs.begin(); oIter != t->recvs.end(); oIter++)
                {
                    std::list<P2POp*>::iterator lIter;
                    for (lIter = oIter->second.begin(); lIter != oIter->second.end(); lIter++)
                    {
                        if (*lIter) (*lIter)->logAsLost(qIter->first);
                    }
                }

                //==== WC-RECVS
                std::list<P2POp*>::iterator lIter;
                for (lIter = t->wcRecvs.begin(); lIter != t->wcRecvs.end(); lIter++)
                {
                    if (*lIter) (*lIter)->logAsLost(qIter->first);
                }
            }//for comms
        }//for processes
	}

#ifdef MUST_DEBUG
	//Print the match histories
	printMatchHistories();
	clearMatchHistories();
	std::cout << "P2PMatch: printed match histories for each comm in files named \"must_match_history_comm_%d.dot\" use DOT to visualize them." << std::endl;
#endif

	//Do a graceful free of all remaining data
	//==Free other data
	//Matching structure
	QT::iterator rankIter;
	for (rankIter = myQs.begin(); rankIter != myQs.end();rankIter++)
	{
	    ProcessQueues::iterator commIter;
	    for (commIter = rankIter->second.begin(); commIter != rankIter->second.end(); commIter++)
	    {
	        //Free the comm
	        commIter->first->erase();

	        //Free any ops in the table
	        std::list<P2POp*>::iterator opIter;

	        //WC Ops
	        for (opIter = commIter->second.wcRecvs.begin(); opIter != commIter->second.wcRecvs.end(); opIter++)
	        {
	            if (*opIter) delete (*opIter);
	        }

	        //Sends
	        RankOutstandings::iterator toIter;
	        for (toIter = commIter->second.sends.begin(); toIter != commIter->second.sends.end(); toIter++)
	        {
	            for (opIter = toIter->second.begin(); opIter != toIter->second.end(); opIter++)
	            {
	                if (*opIter) delete (*opIter);
	            }
	        }

	        //Receives
	        for (toIter = commIter->second.recvs.begin(); toIter != commIter->second.recvs.end(); toIter++)
	        {
	            for (opIter = toIter->second.begin(); opIter != toIter->second.end(); opIter++)
	            {
	                if (*opIter) delete (*opIter);
	            }
	        }
	    }
	}
	myQs.clear ();

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// findMatchingRecv
//=============================
bool P2PMatch::findMatchingRecv (
		/*int from,
		int dest,
		int tag,
		I_Comm* comm,*/
        P2POp* op,
		bool *pOutNeedsToSuspend)
{
	if (pOutNeedsToSuspend)
		*pOutNeedsToSuspend = false;

	//==1) Get the process that would need to call the receive
	QT::iterator q = myQs.find (op->getToRank());
	if (q == myQs.end()) return false;

	//==2) Get the table
	/*
	 * Comm handles may be different on each process,
	 * so we have to ask the comm tracker whether any
	 * of the comms present for the destination are equal
	 * to the given comm.
	 */
	ProcessQueues::iterator t;
	for (t = q->second.begin(); t != q->second.end(); t++)
	{
		if (op->getComm()->compareComms(t->first))
			break;
	}
	if (t == q->second.end()) return false;

	//==3) First search the recvs (they have priority over the wcrecvs)
	RankOutstandings::iterator r = t->second.recvs.find(op->getIssuerRank());

	if (r != t->second.recvs.end())
	{
		std::list<P2POp*>::iterator recv;
		for (recv = r->second.begin(); recv != r->second.end(); recv++)
		{
			if (op->matchTags(*recv))
			{
				//!!! HIT, remove the recv
#ifdef MUST_DEBUG
			    addMatchToHistory (op, *recv);
#endif
			    op->matchTypes (*recv);
			    P2POp *other = *recv;
			    r->second.erase(recv);
			    notifyMatch(op, other);
			    delete (other);
				return true;
			}
		}
	}

	//==4) Search the wcRecvs
	std::list<P2POp*>::iterator recv;
	for (recv = t->second.wcRecvs.begin(); recv != t->second.wcRecvs.end(); recv++)
	{
	    P2POp* other = *recv;
		if (other->getToRank() == op->getToRank() || other->getToRank() == myConsts->getAnySource())
		{
			if (op->matchTags(other))
			{
				if (other->getToRank() != myConsts->getAnySource())
				{
					//!!! HIT, remove the recv
#ifdef MUST_DEBUG
                addMatchToHistory (op, other);
#endif
				    op->matchTypes (*recv);
				    r->second.erase(recv);
				    notifyMatch(op, other);
				    delete (other);
					return true;
				}
				else
				{
					//We need to start suspension
					mySuspendedForEntry = other;

					if (pOutNeedsToSuspend)
						*pOutNeedsToSuspend = true;
					return false;
				}
			}
		}
	}

	return false;
}

//=============================
// addOutstandingSend
//=============================
void P2PMatch::addOutstandingSend (
		/*int from,
		int dest,
		int tag,
		I_Comm* comm,
		bool hasRequest,
		MustRequestType request,
		MustParallelId pId,
		MustLocationId lId*/
        P2POp* op)
{
	//==1) Get the process for the sending process
	QT::iterator q = myQs.find (op->getIssuerRank());
	if (q == myQs.end())
	{
		myQs.insert(std::make_pair(op->getIssuerRank(), ProcessQueues ()));
		q = myQs.find (op->getIssuerRank());
	}

	//==2) Get the table
	ProcessQueues::iterator t = q->second.find(op->getPersistentComm());
	if (t == q->second.end())
	{
		q->second.insert (std::make_pair (op->getCommCopy(), ProcessTable()));
		t = q->second.find(op->getPersistentComm());
	}

	//==3) Get position in sends map
	RankOutstandings::iterator r = t->second.sends.find(op->getToRank());
	if (r == t->second.sends.end())
	{
		t->second.sends.insert (std::make_pair (op->getToRank(), std::list<P2POp*>()));
		r = t->second.sends.find(op->getToRank());
	}

	//==4) Add to queue for the position in sends map
	r->second.push_back(op);
}

//=============================
// findMatchingSend
//=============================
bool P2PMatch::findMatchingSend (
		/*int from,
		int source,
		int tag,
		I_Comm* comm,*/
        P2POp* op,
		bool *pOutNeedsToSuspend)
{
	if (pOutNeedsToSuspend)
		*pOutNeedsToSuspend = false;

	//==1) Get the process(es) that would need to call the send
	bool iterate = false;
	QT::iterator q;

	if (op->getToRank() != myConsts->getAnySource())
	{
		q = myQs.find (op->getToRank());
	}
	else
	{
		q = myQs.begin();
		iterate = true;
	}

	for (;q != myQs.end(); q++)
	{
		//==2) Get the table
		/*
		 * Comm handles may be different on each process,
		 * so we have to ask the comm tracker whether any
		 * of the comms present for the destination is equal
		 * to the given comm.
		 */
		ProcessQueues::iterator t;
		for (t = q->second.begin(); t != q->second.end(); t++)
		{
			if (op->getComm()->compareComms(t->first))
				break;
		}
		if (t == q->second.end())
		{
		    if (iterate)
		        continue;
		    else
		        break;
		}

		//==3) Search the sends
		RankOutstandings::iterator s = t->second.sends.find(op->getIssuerRank());

		if (s != t->second.sends.end())
		{
			std::list<P2POp*>::iterator send;
			for (send = s->second.begin(); send != s->second.end(); send++)
			{
			    P2POp* other = *send;
				if (op->matchTags(other))
				{
					if (op->getToRank() != myConsts->getAnySource())
					{
						//!!! HIT, remove the send
#ifdef MUST_DEBUG
					    addMatchToHistory (other, op);
#endif
					    op->matchTypes (other);
					    s->second.erase(send);
					    notifyMatch(other, op);
                        delete (other);

						return true;
					}
					else
					{
						//!!! Potential hit, we go into suspension now
						if (pOutNeedsToSuspend)
							*pOutNeedsToSuspend = true;
						return false;
					}
				}
			}
		}

		//Force abortion of this loop if we don't need to look into all processes
		if (!iterate)
			break;
	}//For processes to look into

	return false;
}

//=============================
// addOutstandingRecv
//=============================
void P2PMatch::addOutstandingRecv (
		/*int from,
		int source,
		int tag,
		I_Comm* comm,
		bool hasRequest,
		MustRequestType request,
		MustParallelId pId,
		MustLocationId lId*/
        P2POp* op)
{
	//==1) Get the process for the receiving process
	QT::iterator q = myQs.find (op->getIssuerRank());
	if (q == myQs.end())
	{
		myQs.insert(std::make_pair(op->getIssuerRank(), ProcessQueues ()));
		q = myQs.find (op->getIssuerRank());
	}

	//==2) Get the table
	ProcessQueues::iterator t = q->second.find(op->getPersistentComm());
	if (t == q->second.end())
	{
		q->second.insert (std::make_pair (op->getCommCopy(), ProcessTable()));
		t = q->second.find(op->getPersistentComm());
	}

	//==3) Get position in recvs or wcRecvs map
	std::list<P2POp*> *list;

	if (op->getToRank() != myConsts->getAnySource() && t->second.wcRecvs.empty())
	{
		RankOutstandings::iterator r;
		r = t->second.recvs.find(op->getToRank());
		if (r == t->second.recvs.end())
		{
			t->second.recvs.insert (std::make_pair (op->getToRank(), std::list<P2POp*>()));
			r = t->second.recvs.find(op->getToRank());
		}

		list = &(r->second);
	}
	else
	{
		list = &(t->second.wcRecvs);
	}

	//==4) Add to the list
	list->push_back(op);
}

//=============================
// findRecvForUpdate
//=============================
void P2PMatch::findRecvForUpdate (
		int from,
		int newSource,
		bool hasRequest,
		MustRequestType request,
		bool *pOutUnsuspend)
{
    if (pOutUnsuspend) *pOutUnsuspend = false;

	//==1) Search in the tables
	QT::iterator qIter = myQs.find(from);
	if (qIter != myQs.end())
	{
		ProcessQueues::iterator cIter;
		for (cIter = qIter->second.begin(); cIter != qIter->second.end(); cIter++)
		{
			ProcessTable *table = &(cIter->second);

			std::list<P2POp*>::iterator rIter;
			for (rIter = table->wcRecvs.begin(); rIter != table->wcRecvs.end(); rIter++)
			{
				P2POp* entry = *rIter;

				//We are only looking for wc recvs (there may also be other receives in the queue)
				if (entry->getToRank() != myConsts->getAnySource())
					continue;

				//Either both are blocking or both receives are non-blocking
				if (hasRequest != entry->hasRequest())
					continue;

				//If we have a request it must match with the request of the wildcard receive
				if ((hasRequest && entry->getRequest() != request))
					continue;

				//This is the right request:
				//For blocking receives there can always only be one blocking receive
				//in any table or queue, As after that receive it is guaranteed that the
				//update comes before any further blocking wc receive!
				//For non-blocking receives the request is sufficient as identifier.
				//=>Translate new source with given communicator into comm world
				int newSourceTranslated = translateDestSource (entry->getComm(), newSource);
				entry->updateToSource(newSourceTranslated);

				//Determine whether this was the suspension reason
				if ( mySuspendedForEntry && mySuspendedForEntry->hasRequest() == hasRequest &&
						(!hasRequest || mySuspendedForEntry->getRequest() == request) &&
						mySuspendedForEntry->getIssuerRank() == from)
				{
					if (pOutUnsuspend) *pOutUnsuspend = true;
				}

				//Add to the right recv queue
				RankOutstandings::iterator toRecv = table->recvs.find(newSourceTranslated);
				if (toRecv == table->recvs.end())
				{
					table->recvs.insert(std::make_pair(newSourceTranslated, std::list<P2POp*>()));
					toRecv = table->recvs.find(newSourceTranslated);
				}

				toRecv->second.push_back(entry);
				table->wcRecvs.erase(rIter);

				//Look at the wcRecvs and move all non wc receives into their respective
				//receive queues until it is empty or a wc receive is at the front
				while (!table->wcRecvs.empty())
				{
					rIter = table->wcRecvs.begin();

					if ((*rIter)->getToRank() == myConsts->getAnySource())
						break;

					//Add to the right recv queue
					RankOutstandings::iterator toRecv = table->recvs.find((*rIter)->getToRank());
					if (toRecv == table->recvs.end())
					{
						table->recvs.insert(std::make_pair((*rIter)->getToRank(), std::list<P2POp*>()));
						toRecv = table->recvs.find((*rIter)->getToRank());
					}

					toRecv->second.push_back(*rIter);
					table->wcRecvs.erase(rIter);
				}

				//Done
				return;
			}
		}
	}

	//==2) Search in the suspended queue
	std::map<int, std::list<P2POp* > >::iterator wcRQueue = suspendedWcRecvs.find(from);
	if (wcRQueue == suspendedWcRecvs.end())
		return; // TODO should warn that there was an update without a recv to update

	std::list<P2POp*>::iterator susIter;

	for (susIter = wcRQueue->second.begin(); susIter != wcRQueue->second.end(); susIter++)
	{
		P2POp* wc = *susIter;

		//Request property must match
		if (wc->hasRequest() != hasRequest)
			continue;

		//If there is an associated request it must match
		if (hasRequest && wc->getRequest() != request)
			continue;

		//Hit! Update ! (No translation needed, it was not yet processed and will be translated at that time)
		wc->updateToSource(translateDestSource (wc->getComm(), newSource));

		wcRQueue->second.erase(susIter);

		//Determine whether this was the suspension reason
		if (	mySuspendedForEntry && mySuspendedForEntry->hasRequest() == hasRequest &&
				(!hasRequest || mySuspendedForEntry->getRequest() == request) &&
				mySuspendedForEntry->getIssuerRank() == from)
		{
			if (pOutUnsuspend) *pOutUnsuspend = true;
		}

		break;
	}
}

//=============================
// printQs
//=============================
void P2PMatch::printQs (void)
{
	std::cout << "====== Qs ======" << std::endl;

	QT::iterator qIter;
	for (qIter = myQs.begin(); qIter != myQs.end(); qIter++)
	{
		std::cout << qIter->first << ":" << std::endl;

		ProcessQueues::iterator cIter;
		for (cIter = qIter->second.begin(); cIter != qIter->second.end(); cIter++)
		{
			std::cout << "    " << cIter->first << ":" << std::endl;

			ProcessTable *t = &(cIter->second);

			RankOutstandings::iterator oIter;

			//==== SENDS
			std::cout << "        Sends:" << std::endl;
			for (oIter = t->sends.begin(); oIter != t->sends.end(); oIter++)
			{
				std::cout << "            To " << oIter->first << ": " << std::endl;

				std::list<P2POp*>::iterator lIter;
				for (lIter = oIter->second.begin(); lIter != oIter->second.end(); lIter++)
				{
				    std::cout << "                ";
					if(*lIter) (*lIter)->print(std::cout);
					std::cout << std::endl;
				}

				std::cout << std::endl;
			}

			//==== RECVS
			std::cout << "        Recvs:" << std::endl;
			for (oIter = t->recvs.begin(); oIter != t->recvs.end(); oIter++)
			{
				std::cout << "            From " << oIter->first << ": " << std::endl;

				std::list<P2POp*>::iterator lIter;
				for (lIter = oIter->second.begin(); lIter != oIter->second.end(); lIter++)
				{
				    std::cout << "                ";
				    if(*lIter) (*lIter)->print(std::cout);
				    std::cout << std::endl;
				}

				std::cout << std::endl;
			}

			//==== WC-RECVS
			std::cout << "        wcRecvs: " << std::endl;
			std::list<P2POp*>::iterator lIter;
			for (lIter = t->wcRecvs.begin(); lIter != t->wcRecvs.end(); lIter++)
			{
			    std::cout << "                ";
			    if(*lIter) (*lIter)->print(std::cout);
			    std::cout << std::endl;
			}
			std::cout << std::endl;

		}//for comms
	}//for processes

	std::cout << "====== Suspended WC-Receives ======" << std::endl;
	std::map<int, std::list<P2POp*> >::iterator rankWcIter;
	for (rankWcIter = suspendedWcRecvs.begin(); rankWcIter != suspendedWcRecvs.end(); rankWcIter++)
	{
	    std::cout << "RANK " << rankWcIter->first << ":" << std::endl;

	    std::list<P2POp*>::iterator sus;
		for (sus = rankWcIter->second.begin(); sus != rankWcIter->second.end(); sus++)
		{
		    P2POp* entry = *sus;

			if (entry)
				std::cout << "    ->";

			entry->print(std::cout);
			std::cout << std::endl;
		}
	}
}

//=============================
// addMatchToHistory
//=============================
#ifdef MUST_DEBUG
void P2PMatch::addMatchToHistory (P2POp* send, P2POp* recv)
{
    //Add to matches for graphical representation
    AllMatchHistories::iterator historyIter;
    for (historyIter = myMatches.begin(); historyIter != myMatches.end(); historyIter++)
        if (historyIter->first->compareComms(send->getComm()))
            break;
    if (historyIter == myMatches.end())
    {
        std::pair<AllMatchHistories::iterator, bool> ret = myMatches.insert (std::make_pair(send->getCommCopy(), MatchHistory ()));
        historyIter = ret.first;
    }

    historyIter->second[send->getIssuerRank()][recv->getIssuerRank()].push_back (send->getTag());
}
#endif

//=============================
// printMatchHistories
//=============================
#ifdef MUST_DEBUG
void P2PMatch::printMatchHistories (void)
{
    /**
     * @todo with intercommunicators the comm translation will be very wacky, think about that.
     */

    AllMatchHistories::iterator cIter;
    int d = 0;
    for (cIter = myMatches.begin(); cIter != myMatches.end(); cIter++, d++)
    {
        I_Comm *comm = cIter->first;

        MatchHistory::iterator sendIter;

        std::stringstream stream, infostr;
        stream << "must_match_history_comm_" << d << ".dot";
        std::ofstream out (stream.str().c_str());
        std::list<std::pair <MustParallelId, MustLocationId> > references;

        out << "digraph MsgMatchHistory {" << std::endl;
        out << "commInfo [label=\"";
        comm->printInfo (infostr, &references);
        out << infostr.str() << "\"];"<<std::endl;

        for (sendIter = cIter->second.begin(); sendIter != cIter->second.end(); sendIter++)
        {
            int send = sendIter->first;
            int sendTranslated = invTranslateDestSource (comm, send);

            std::map<int, std::list<int> >::iterator receiveIter;
            for (receiveIter = sendIter->second.begin(); receiveIter != sendIter->second.end(); receiveIter++)
            {
                int receive = receiveIter->first;
                int receiveTranslated = invTranslateDestSource (comm, receive);

                out << sendTranslated << "->" << receiveTranslated << " [label=\"";

                std::list<int>::iterator tagIter;
                for (tagIter = receiveIter->second.begin(); tagIter != receiveIter->second.end(); tagIter++)
                {
                    int tag = *tagIter;

                    if (tagIter != receiveIter->second.begin())
                        out << ", ";

                    out << tag;
                }

                out << "\"];" << std::endl;
            }
        }

        out << "}" << std::endl;
    }
}
#endif

//=============================
// clearMatchHistories
//=============================
#ifdef MUST_DEBUG
void P2PMatch::clearMatchHistories(void)
{
    AllMatchHistories::iterator cIter;
    for (cIter = myMatches.begin(); cIter != myMatches.end(); cIter++)
    {
        cIter->first->erase();
    }
    myMatches.clear ();
}
#endif

//=============================
// registerListener
//=============================
GTI_RETURN P2PMatch::registerListener (I_P2PMatchListener *listener)
{
    myListeners.push_back (listener);
    return GTI_SUCCESS;
}

//=============================
// notifyMatch
//=============================
void P2PMatch::notifyMatch (P2POp* send, P2POp *recv)
{
    std::list<I_P2PMatchListener*>::iterator lIter;
    for (lIter = myListeners.begin(); lIter != myListeners.end(); lIter++)
    {
        I_P2PMatchListener* listener = *lIter;
        listener->newMatch(
                send->getIssuerRank(),
                recv->getIssuerRank(),
                send->hasRequest(),
                send->getRequest(),
                recv->hasRequest(),
                recv->getRequest(),
                send->getSendMode());
    }
}

//=============================
// getP2PInfo
//=============================
bool P2PMatch::getP2PInfo (int rank, MustRequestType request, P2PInfo* outInfo)
{
    QT::iterator pos = myQs.find(rank);

    if (pos == myQs.end()) return false;

    //We don't know the comm, lets try all of them!
    ProcessQueues::iterator cIter;
    for (cIter = pos->second.begin(); cIter != pos->second.end(); cIter++)
    {
        ProcessTable *table = &(cIter->second);

        //Sends & Receives
        RankOutstandings::iterator toIter;
        for (int i = 0; i < 2; i++)
        {
            RankOutstandings *outstandings = &(table->sends);
            if (i == 1)
                outstandings = &(table->recvs);

            for (toIter = outstandings->begin(); toIter != outstandings->end(); toIter++)
            {
                std::list<P2POp*>::iterator opIter;
                for (opIter = toIter->second.begin(); opIter != toIter->second.end(); opIter++)
                {
                    P2POp* op = *opIter;

                    if (!op->hasRequest()) continue;

                    if (op->getRequest() != request) continue;

                    fillInfo (op, outInfo);
                    return true;
                }
            }
        }

        //Wc receives
        std::list<P2POp*>::iterator opIter;
        for (opIter = table->wcRecvs.begin(); opIter != table->wcRecvs.end(); opIter++)
        {
            P2POp* op = *opIter;

            if (!op->hasRequest()) continue;

            if (op->getRequest() != request) continue;

            fillInfo (op, outInfo);
            return true;
        }
    }//For communicators

    return false;
}

//=============================
// getP2PInfo
//=============================
bool P2PMatch::getP2PInfo (int rank, bool isSend, P2PInfo* outInfo)
{
    QT::iterator pos = myQs.find(rank);

    if (pos == myQs.end()) return false;

    //We don't know the comm, lets try all of them!
    ProcessQueues::iterator cIter;
    for (cIter = pos->second.begin(); cIter != pos->second.end(); cIter++)
    {
        ProcessTable *table = &(cIter->second);

        //Sends & Receives
        RankOutstandings::iterator toIter;
        RankOutstandings *outstandings = &(table->sends);
        if (!isSend)
            outstandings = &(table->recvs);

        for (toIter = outstandings->begin(); toIter != outstandings->end(); toIter++)
        {
            std::list<P2POp*>::iterator opIter;
            for (opIter = toIter->second.begin(); opIter != toIter->second.end(); opIter++)
            {
                P2POp* op = *opIter;

                if (op->hasRequest()) continue;

                //IMPORTANT: Ignore "blocking" bsends
                if (op->isSend() && op->getSendMode() == MUST_BUFFERED_SEND) continue;

                fillInfo (op, outInfo);
                return true;
            }
        }

        //Wc receives
        if (!isSend)
        {
            std::list<P2POp*>::iterator opIter;
            for (opIter = table->wcRecvs.begin(); opIter != table->wcRecvs.end(); opIter++)
            {
                P2POp* op = *opIter;

                if (op->hasRequest()) continue;

                //IMPORTANT: Ignore "blocking" bsends
                if (op->isSend() && op->getSendMode() == MUST_BUFFERED_SEND) continue;

                fillInfo (op, outInfo);
                return true;
            }
        }//If a receive is being looked for
    }//For communicators

    return false;
}

//=============================
// fillInfo
//=============================
void P2PMatch::fillInfo (P2POp* op, P2PInfo *info)
{
    if (!info || !op)
        return;

    info->comm = op->getComm();
    info->isSend = op->isSend ();
    info->isWc = op->wasIssuedAsWcReceive();
    info->lId = op->getLId();
    info->pId = op->getPId();
    info->mode = op->getSendMode();
    info->target = op->getToRank();
}

//=============================
// notifyDeadlock
//=============================
void P2PMatch::notifyDeadlock (void)
{
    myPrintLostMessages = false;
}

//=============================
// checkpoint
//=============================
void P2PMatch::checkpoint (void)
{
    //TODO
}

//=============================
// rollback
//=============================
void P2PMatch::rollback (void)
{
    //TODO
}

/*EOF*/
