/*
 * MUST (Marmot Umpire Scalable Tool)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2010-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2010-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file LocationImpl.cpp
 *       Implementation for the location id analysis interface.
 *
 *  @date 07.01.2010
 *  @author Tobias Hilbrich
 */

#include "GtiMacros.h"

#include "LocationImpl.h"

using namespace must;

mGET_INSTANCE_FUNCTION(LocationImpl);
mFREE_INSTANCE_FUNCTION(LocationImpl);
mPNMPI_REGISTRATIONPOINT_FUNCTION(LocationImpl);

//=============================
// Constructor
//=============================
LocationImpl::LocationImpl (const char* instanceName)
    : gti::ModuleBase<LocationImpl, I_LocationAnalysis> (instanceName),
      myGlobalMap (),
      myLocalMap (),
      myEmptyInfo ()
{
    //create sub modules
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();
    //No sub modules needed ...

    myEmptyInfo.callName = "";
}

//=============================
// Destructor
//=============================
LocationImpl::~LocationImpl ()
{
	myGlobalMap.clear ();
	myLocalMap.clear ();
}

//=============================
// registerLocation
//=============================
GTI_ANALYSIS_RETURN LocationImpl::registerLocation (
	MustParallelId pId,
	MustLocationId lId,
	char* callName,
	int callNameLen
#ifdef USE_CALLPATH
	,
	int numStackLevels,
	int stackInfosLength,
	int indicesLength,
	int* infoIndices,
	char* stackInfos
#endif
    )
{
#ifdef MUST_DEBUG
		std::cout << "DEBUG: Added location with id " << lId << " from parallel id " << pId << std::endl;
#endif

	//===Determine whether a local mapping for this location is present
	std::pair<MustParallelId, MustLocationId> idPair = std::make_pair (pId, lId);
	LocalMapType::iterator lPos = myLocalMap.find (idPair);

	if (lPos != myLocalMap.end())
		//already an entry present, nothing todo (we silently assume that the infos are both the same, hopefully this is usually true
		return GTI_ANALYSIS_SUCCESS;

	//===Build the location info
	LocationInfo info;
	info.callName = callName;

#ifdef USE_CALLPATH
	int level;
	int index = 0;
	for (level = 0; level < numStackLevels; level++)
	{
	    MustStackLevelInfo levelInfo;
	    for (int sub = 0; sub < 3; sub++)
	    {
	        int max = infoIndices[level*3 + sub];

	        char temp[1024];

	        int curr = 0;
	        while (index <= max && curr < 1024)
	        {
	            temp[curr] = stackInfos[index];
	            index++;
	            curr++;
	        }
	        temp[curr] = '\0'; //End (not necessary unleas larger 1024)

	        switch (sub)
	        {
	        case 0:
	            levelInfo.symName = temp;
	            break;
	        case 1:
	            levelInfo.fileModule = temp;
	            break;
	        case 2:
	            levelInfo.lineOffset = temp;
	            break;
	        }
	    }

	    info.stack.push_back (levelInfo);

	    //DEBUG
	    //std::cout << "Location:" <<  info.callName << " -> " << levelInfo.symName << "@" << levelInfo.fileModule << ":" << levelInfo.lineOffset << std::endl;
	}
#endif

	//===Is an equal global mapping already present ?
	GlobalMapType::iterator gPos = myGlobalMap.find (lId);

	if (gPos != myGlobalMap.end())
	{
	    bool addToLocal = false;

		//==Already present, if it differs from the current entry we need to add it to local
		LocationInfo other = gPos->second;
		if (other.callName != info.callName)
			addToLocal = true;

#ifdef USE_CALLPATH
		if (other.stack.size () != info.stack.size())
		    addToLocal = true;

		if (!addToLocal)
		{
		    std::list<MustStackLevelInfo>::iterator a, b;
		    for (a = other.stack.begin(), b = info.stack.begin(); a != other.stack.end(); a++, b++)
		    {
		        if (a->symName != b->symName || a->lineOffset != b->lineOffset)
		        {
		            addToLocal = true;
		            break;
		        }
		    }
		}
#endif

		if (addToLocal)
		{
		    myLocalMap.insert (std::make_pair (idPair, info));
		}
	}
	else
	{
		//Not in gobal yet -> add to global
		myGlobalMap.insert (std::make_pair(lId, info));
	}

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// getInfoForId
//=============================
LocationInfo& LocationImpl::getInfoForId (MustParallelId pId, MustLocationId lId)
{
	//=== Is a local entry present -> use it
	std::pair<MustParallelId, MustLocationId> idPair = std::make_pair (pId, lId);
	LocalMapType::iterator lPos = myLocalMap.find (idPair);
	if (lPos != myLocalMap.end ())
		return lPos->second;

	//=== Is a global entry present -> use it
	GlobalMapType::iterator gPos = myGlobalMap.find (lId);
	if (gPos != myGlobalMap.end())
		return gPos->second;

	//=== Ups, unknown location id, we return something in that case
	//        Probably empty string for call name is a good indicator for that.
	////We currently do not warn about this, as the location reduction will actually lead to such cases
	////If we still want to warn we must use an extra interface call for the location reduction
	//static bool unknownWarned = false;
	//if (!unknownWarned)
	//	std::cerr << "Unknown location with id " << lId << " (parallel id " << pId << ") (This warning is only given once)" << std::endl;
	//unknownWarned = true;

	return myEmptyInfo;
}

//=============================
// toString
//=============================
std::string LocationImpl::toString (MustParallelId pId, MustLocationId lId)
{
	LocationInfo info = getInfoForId (pId, lId);
	std::string ret = "";

	ret += "call " + info.callName;

	return ret;
}

/*EOF*/
