/*
 * MUST (Marmot Umpire Scalable Tool)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2011-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2011-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file OverlapChecks.cpp
 *       @see must::OverlapChecks.
 *
 *  @date 27.05.2011
 *  @author Joachim Protze
 */

#include "GtiMacros.h"
#include "OverlapChecks.h"
#include "MustEnums.h"

#include <sstream>

using namespace must;

mGET_INSTANCE_FUNCTION(OverlapChecks);
mFREE_INSTANCE_FUNCTION(OverlapChecks);
mPNMPI_REGISTRATIONPOINT_FUNCTION(OverlapChecks);


int OverlapChecks::pId2Rank(
        MustParallelId pId)
{
    return myPIdMod->getInfoForId(pId).rank;
}

//=============================
// Constructor.
//=============================
OverlapChecks::OverlapChecks (const char* instanceName)
    : ModuleBase<OverlapChecks, I_OverlapChecks> (instanceName),
    activeBlocks (),
    preparedBlocklists (),
    activeRequestsBlocklists ()
{
    //create sub modules
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //handle sub modules
#define NUM_SUBMODULES 5
    if (subModInstances.size() < NUM_SUBMODULES)
    {
            std::cerr << "Module has not enough sub modules, check its analysis specification! (" << __FILE__ << "@" << __LINE__ << ")" << std::endl;
            assert (0);
    }
    if (subModInstances.size() > NUM_SUBMODULES)
    {
            for (int i = NUM_SUBMODULES; i < subModInstances.size(); i++)
                destroySubModuleInstance (subModInstances[i]);
    }

    myPIdMod = (I_ParallelIdAnalysis*) subModInstances[0];
    myLogger = (I_CreateMessage*) subModInstances[1];
    myArgMod = (I_ArgumentAnalysis*) subModInstances[2];
    myDatMod = (I_DatatypeTrack*) subModInstances[3];
    myReqMod = (I_RequestTrack*) subModInstances[4];

    //Initialize module data
    //Nothing to do
}

//=============================
// Destructor.
//=============================
OverlapChecks::~OverlapChecks (void)
{
    if (myPIdMod)
        destroySubModuleInstance ((I_Module*) myPIdMod);
    myPIdMod = NULL;

    if (myLogger)
        destroySubModuleInstance ((I_Module*) myLogger);
    myLogger = NULL;

    if (myArgMod)
        destroySubModuleInstance ((I_Module*) myArgMod);
    myArgMod = NULL;

    if (myDatMod)
        destroySubModuleInstance ((I_Module*) myDatMod);
    myDatMod = NULL;

    if (myReqMod)
        destroySubModuleInstance ((I_Module*) myReqMod);
    myReqMod = NULL;


#ifdef MUST_DEBUG
    int myCount=0;
    for (mustPidMap< MustMemIntervalListType >::iterator iter1=activeBlocks.begin(); iter1!=activeBlocks.end(); iter1++)
            myCount+=iter1->second.size();
    std::cout << "Sizeof(activeBlocks) = " << myCount << std::endl;

    myCount=0;
    for (mustPidRequestMap< MustMemIntervalListType >::iterator iter1=preparedBlocklists.begin(); iter1!=preparedBlocklists.end(); iter1++)
        for (std::map< MustRequestType, MustMemIntervalListType >::iterator iter2=iter1->second.begin(); iter2!=iter1->second.end(); iter2++)
            myCount+=iter2->second.size();
    std::cout << "Sizeof(preparedBlocklists) = " << myCount << std::endl;

    myCount=0;
    for (mustPidRequestMap< std::list< MustMemIntervalListType::iterator > >::iterator iter1=activeRequestsBlocklists.begin(); iter1!=activeRequestsBlocklists.end(); iter1++)
        for (std::map< MustRequestType, std::list< MustMemIntervalListType::iterator > >::iterator iter2=iter1->second.begin(); iter2!=iter1->second.end(); iter2++)
            myCount+=iter2->second.size();
    std::cout << "Sizeof(activeRequestsBlocklists) = " << myCount << std::endl;
#endif

}


GTI_ANALYSIS_RETURN OverlapChecks::sendRecvMatches (
        MustParallelId pId,
        MustLocationId lId,
        int sendcount,
        MustDatatypeType sendtype,
        int recvcount,
        MustDatatypeType recvtype
        )
{
    MustAddressType errorpos;
    I_Datatype * typeinfoA = myDatMod->getDatatype(pId, sendtype);
    I_Datatype * typeinfoB = myDatMod->getDatatype(pId, recvtype);
    MustMessageIdNames ret;
    if (typeinfoA!=NULL && typeinfoB!= NULL)
        ret = typeinfoA->isSubsetOfB(sendcount, typeinfoB, recvcount, &errorpos);
    else
        ret = MUST_ERROR_TYPEMATCH_INTERNAL_NOTYPE;
    if (ret == MUST_MESSAGE_NO_ERROR)
        return GTI_ANALYSIS_SUCCESS;

    std::stringstream stream;
    switch (ret){
        case MUST_ERROR_TYPEMATCH_MISSMATCH:
            stream
                << "type missmatch in operation after " << errorpos << " bytes" << std::endl;
            break;
        case MUST_ERROR_TYPEMATCH_MISSMATCH_BYTE:
            stream
                << "type missmatch in operation after " << errorpos << " bytes" << std::endl;
            break;
        case MUST_ERROR_TYPEMATCH_INTERNAL_NOTYPE:
            stream
                << "internal error in operation: error on getting typeinfo" << std::endl;
            break;
        case MUST_ERROR_TYPEMATCH_INTERNAL_TYPESIG:
            stream
                << "internal error in operation: error on getting typesig" << std::endl;
            break;
        case MUST_ERROR_TYPEMATCH_LENGTH:
            stream
                << "length missmatch in operation after " << errorpos << " bytes" << std::endl;
            break;
    }

    myLogger->createMessage(
            ret,
            pId,
            lId,
            MustErrorMessage,
            stream.str()
            );
    return GTI_ANALYSIS_FAILURE;
}

//=============================
//  Helper function for isTypeOverlapped(N)
//=============================

bool OverlapChecks::checkTypeOverlapped (
        I_Datatype * typeinfo,
        int extent,
        int count)
{
    bool overlapped;
    if (typeinfo==NULL)
        return false;
    BlockInfo& blockInfo = typeinfo->getBlockInfo();
    MustStridedBlocklistType blocklist = blockInfo.stridedBlocklist;
    if (blockInfo.overlapped)
        return true;
    if (count > 1) // if we want to check copys, we must duplicate the blocklist first
    {
        MustStridedBlocklistType tempList;
        MustStridedBlocklistType::iterator iter, insertIter=tempList.begin();
        int i;
        for (iter = blocklist.begin(); iter != blocklist.end(); iter++)
        {
            for (i=0; i<count; i++)
            {
                insertIter = tempList.insert(insertIter, StridedBlock(*iter, extent * i, i));
            }
        }
//         tempList.sort();
        blocklist=tempList;
    }
    MustStridedBlocklistType::iterator iter, nextIter;

    // run over the (sorted!) blocklist and check neighbors for overlaps
    if (isOverlapped(blocklist, iter, nextIter))
    {
        std::cout << "selfoverlapping buffer on " << iter->repetition << "(th) repetition at ";
        typeinfo->printDatatypeLongPos(std::cout, iter->pos + iter->first - nextIter->first);
        std::cout << " with " << nextIter->repetition  << "(th) repetition at ";
        typeinfo->printDatatypeLongPos(std::cout, nextIter->pos);
        std::cout << std::endl;
        return true;
    }
    return false;
}


//=============================
// isOverlappedN
//=============================
bool OverlapChecks::isOverlappedN(
        MustParallelId pId,
        MustLocationId lId,
        MustDatatypeType datatype,
        int count)
{
    I_Datatype * typeinfo = myDatMod->getDatatype(pId, datatype);
    if (typeinfo==NULL){
        return false;
    }
    std::pair<int,int> cachevalues = typeinfo->getSelfoverlapCache();
    if (count<=cachevalues.first)
        return false;
    if (count>=cachevalues.second)
        return true;
    
    MustAddressType extent = typeinfo->getExtent(),
                    true_extent = typeinfo->getTrueExtent();
    int myCount = 1;
    if (count > 1 && extent < true_extent) // copys can only overlap, if extent < true_extent
    {
        // filling the first copy of the datatype (width = 'true_extent')
        // with 'myCount' copies , so that 'myCount' * 'extent' > 'true_extent'
        // -> the start of the (myCount+1)'th copy is after the end of the first copy

        // calculate myCount
        myCount = true_extent / extent + 1;

        // myCount = min(count, myCount)
        // we dont want to do more than count or myCount
        if (count < myCount)
            myCount = count;
    }
    if (myCount<=cachevalues.first)
        return false;
    return setSelfOverlapCache(typeinfo, myCount, checkTypeOverlapped(typeinfo, extent, myCount));
}

//=============================
// set entry to overlapCache
//=============================

bool OverlapChecks::setSelfOverlapCache(
        I_Datatype * typeinfo,
        int count,
        bool overlaps
        )
{
    if (overlaps)
        typeinfo->setMinOverlap(count);
    else
        typeinfo->setMaxNoOverlap(count);
    return overlaps;
}


//=============================
// isSendRecvOverlapped
//=============================
GTI_ANALYSIS_RETURN OverlapChecks::isSendRecvOverlapped (
        MustParallelId pId,
        MustLocationId lId,
        MustAddressType sendbuf,
        int sendcount,
        MustDatatypeType sendtype,
        MustAddressType recvbuf,
        int recvcount,
        MustDatatypeType recvtype
        )
{
    return isSendRecvOverlappedN(pId, lId, sendbuf, NULL, 0, &sendcount, 1, &sendtype, 1, recvbuf, NULL, 0, &recvcount, 1, &recvtype, 1);
}

//=============================
// isSendRecvOverlappedN
//=============================
GTI_ANALYSIS_RETURN OverlapChecks::isSendRecvOverlappedN (
        MustParallelId pId,
        MustLocationId lId,
        MustAddressType sendbuf,
        int *senddispls,
        int senddisplslen,
        int *sendcounts,
        int sendcountslen,
        MustDatatypeType *sendtypes,
        int sendtypeslen,
        MustAddressType recvbuf,
        int *recvdispls,
        int recvdisplslen,
        int *recvcounts,
        int recvcountslen,
        MustDatatypeType *recvtypes,
        int recvtypeslen
        )
{
    if (sendcountslen<1 || sendtypeslen<1 || recvcountslen<1 || recvtypeslen<1 || senddisplslen<0 || recvdisplslen<0 || (senddisplslen>0 && senddispls==NULL) || (recvdisplslen>0 && recvdispls==NULL))
    {
        // something went really wrong, exit test
        std::cout << "Implementation error: Malicious call of OverlapChecks::isSendRecvOverlappedN !" << std::endl;
        return GTI_ANALYSIS_SUCCESS; 
    }
    MustMemIntervalListType iList, tList;
    int displacement=0, count = sendcounts[0];
    MustDatatypeType type = sendtypes[0];
    I_Datatype * typeinfo = myDatMod->getDatatype(pId, type);
    if (typeinfo == NULL){
//        std::cout << "Datatype " << type << "not found in isSendRecvOverlappedN" << std::endl;
        return GTI_ANALYSIS_SUCCESS;
    }
    MustAddressType extent = typeinfo->getExtent();
    // get all blocks, used for communication
    // at first blocks from the send part
    for (int i=0; i<sendcountslen; i++)
    {
        if (senddisplslen>1) // we have displacements
            displacement = senddispls[i];
        if (sendcountslen>1) // we have counts
            count = sendcounts[i];
        if (sendtypeslen>1) // we have types, this means bytewise displacements
        {
            type = sendtypes[i];
            typeinfo = myDatMod->getDatatype(pId, type);
            if (typeinfo == NULL)
                return GTI_ANALYSIS_SUCCESS;
        }
        else // we have one type, so displacements are given in extent-width
            displacement *= extent;
        tList = calcIntervalList(typeinfo, sendbuf + displacement, count, (MustRequestType)0, true);
        iList.insert(tList.begin(),tList.end());
    }

    // till now overlaps were harmless and allowed
    bool isSelfoverlaping = false;
    // from now on an overlap is harmful and results in an error

    type = recvtypes[0];
    displacement = 0;
    count = recvcounts[0];
    typeinfo = myDatMod->getDatatype(pId, type);
    if (typeinfo == NULL)
        return GTI_ANALYSIS_SUCCESS;
    extent = typeinfo->getExtent();
    for (int i=0; i<recvcountslen && !isSelfoverlaping; i++)
    {
        if (recvdisplslen>1) // we have displacements
            displacement = recvdispls[i];
        if (recvcountslen>1) // we have counts
            count = recvcounts[i];
        if (recvtypeslen>1) // we have types, this means bytewise displacements
            type = recvtypes[i];
            typeinfo = myDatMod->getDatatype(pId, type);
            if (typeinfo == NULL)
                return GTI_ANALYSIS_SUCCESS;
        else // we have one type, so displacements are given in extent-width
            displacement *= extent;
        tList = calcIntervalList(typeinfo, recvbuf + displacement, count, (MustRequestType)0, false, &isSelfoverlaping);
        iList.insert(tList.begin(), tList.end());
    }
    if (isSelfoverlaping)
    {
        // selfoverlaping has been seen with other test!
        return GTI_ANALYSIS_SUCCESS;
    }

    MustMemIntervalListType::iterator iter, nextIter;

    // selfoverlaping -> error !
    if (isOverlapped(iList, iter, nextIter, true))
    {
        std::stringstream stream;
        // selfoverlap is found before!
/*        if (iter->type == nextIter->type)
            return GTI_ANALYSIS_SUCCESS;*/
        
        MustMemIntervalListType::iterator recvIter = (!iter->isSend) ? iter : nextIter,
                                          sendIter = (!iter->isSend) ? nextIter : iter;
        stream
            <<  "The memory regions spanned by the recv part overlaps at the "
            <<  recvIter->repetition << "(th) repetition of datatype at its position ";
        recvIter->type->printDatatypeLongPos(stream, recvIter->pos + nextIter->first - recvIter->first);
        stream
            << " with regions spanned by the ";
        if (!sendIter->merged)
        {
            stream << sendIter->repetition << "(th) repetition of datatype at its position ";
            sendIter->type->printDatatypeLongPos(stream, sendIter->pos + nextIter->first - sendIter->first);
            stream << " used in ";
        }
        if (sendIter->isSend)
            stream << "send";
        else
            stream << "receive";
        stream << " part of this operation!" << std::endl;
//         else
//             return GTI_ANALYSIS_SUCCESS;

        myLogger->createMessage(
                MUST_ERROR_SELFOVERLAPPED,
                pId,
                lId,
                MustErrorMessage,
                stream.str()
                );
        return GTI_ANALYSIS_FAILURE;
    }
    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// isTypeOverlapped
//=============================
GTI_ANALYSIS_RETURN OverlapChecks::isTypeOverlapped (
        MustParallelId pId,
        MustLocationId lId,
        int aId,
        MustDatatypeType datatype)
{
    return isTypeOverlappedN(pId, lId, aId, datatype, 1);
}

//=============================
// isTypeOverlappedN
//=============================
GTI_ANALYSIS_RETURN OverlapChecks::isTypeOverlappedN (
        MustParallelId pId,
        MustLocationId lId,
        int aId,
        MustDatatypeType datatype,
        int count)
{
    if (isOverlappedN(pId, lId, datatype, count))
    {
        std::list<std::pair<MustParallelId,MustLocationId> > refs;
        std::stringstream stream;
        stream
            <<  "Argument " << myArgMod->getIndex(aId) << " (" << myArgMod->getArgName(aId)
            << ") is selfoverlapping after repetition of " << count << " !" << std::endl << "Infos on Datatype:";
            
        myDatMod->getDatatype(pId, datatype)->printInfo(stream, &refs);

        myLogger->createMessage(
                MUST_ERROR_SELFOVERLAPPED,
                pId,
                lId,
                MustErrorMessage,
                stream.str(), 
                refs
                );
        return GTI_ANALYSIS_FAILURE;
    }

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// warnIfTypeOverlapped
//=============================
GTI_ANALYSIS_RETURN OverlapChecks::warnIfTypeOverlapped (
        MustParallelId pId,
        MustLocationId lId,
        int aId,
        MustDatatypeType datatype)
{
    if (isOverlappedN(pId, lId, datatype, 1))
    {
        std::list<std::pair<MustParallelId,MustLocationId> > refs;
        std::stringstream stream;
        stream
            <<  "Argument " << myArgMod->getIndex(aId) << " (" << myArgMod->getArgName(aId)
            << ") is selfoverlapping !" << std::endl << "Infos on Datatype:";
            
        myDatMod->getDatatype(pId, datatype)->printInfo(stream, &refs);

        myLogger->createMessage(
                MUST_WARNING_SELFOVERLAPPED,
                pId,
                lId,
                MustWarningMessage,
                stream.str(), 
                refs
                );
        return GTI_ANALYSIS_FAILURE;
    }

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// warnIfTypeOverlapped
//=============================
GTI_ANALYSIS_RETURN OverlapChecks::warnIfTypeOverlappedN (
        MustParallelId pId,
        MustLocationId lId,
        int aId,
        MustDatatypeType datatype,
        int count)
{
    if (isOverlappedN(pId, lId, datatype, count))
    {
        std::list<std::pair<MustParallelId,MustLocationId> > refs;
        std::stringstream stream;
        stream
            <<  "Argument " << myArgMod->getIndex(aId) << " (" << myArgMod->getArgName(aId)
            << ") is selfoverlapping after repetition of " << count << " !" << std::endl << "Infos on Datatype:";
            
        myDatMod->getDatatype(pId, datatype)->printInfo(stream, &refs);

        myLogger->createMessage(
                MUST_WARNING_SELFOVERLAPPED,
                pId,
                lId,
                MustWarningMessage,
                stream.str(), 
                refs
                );
        return GTI_ANALYSIS_FAILURE;
    }

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// isRecvOverlappedN
//=============================
GTI_ANALYSIS_RETURN OverlapChecks::isRecvOverlappedN (
        MustParallelId pId,
        MustLocationId lId,
        MustDatatypeType datatype,
        int count)
{
    if (isOverlappedN(pId, lId, datatype, count))
    {
        std::stringstream stream;
        stream
            <<  "Datatype used for receive is selfoverlapping after repetition of " << count << " !" << std::endl;

        myLogger->createMessage(
                MUST_ERROR_SELFOVERLAPPED,
                pId,
                lId,
                MustErrorMessage,
                stream.str()
                );
        return GTI_ANALYSIS_FAILURE;
    }

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// isSendOverlappedN
//=============================
GTI_ANALYSIS_RETURN OverlapChecks::isSendOverlappedN (
        MustParallelId pId,
        MustLocationId lId,
        MustDatatypeType datatype,
        int count)
{
    if (isOverlappedN(pId, lId, datatype, count))
    {
        std::stringstream stream;
        stream
            <<  "Datatype used for send is selfoverlapping after repetition of " << count << " !" << std::endl;

        myLogger->createMessage(
                MUST_WARNING_SELFOVERLAPPED,
                pId,
                lId,
                MustWarningMessage,
                stream.str()
                );
        return GTI_ANALYSIS_FAILURE;
    }

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// sendOverlapcheckCounts
//=============================
GTI_ANALYSIS_RETURN OverlapChecks::sendOverlapcheckCounts (
        MustParallelId pId,
        MustLocationId lId,
        MustAddressType buffer,
        int displs[],
        int counts[],
        MustDatatypeType datatype,
        int commsize
        )
{
    MustMemIntervalListType iList, tList;
    bool isSelfoverlaping = false;
    I_Datatype * typeinfo = myDatMod->getDatatype(pId, datatype);
    if (typeinfo == NULL)
        return GTI_ANALYSIS_SUCCESS;
    MustAddressType extent = typeinfo->getExtent();
    // get all blocks, used for communication
    for (int i=0; i<commsize; i++)
    {
        tList = calcIntervalList(typeinfo, buffer+displs[i]*extent, counts[i], (MustRequestType)0, true, &isSelfoverlaping);
        iList.insert(tList.begin(), tList.end());
    }

    MustMemIntervalListType::iterator iter, nextIter;

    // selfoverlaping -> warning for send, finish check
    if (isOverlapped(iList, iter, nextIter, false))
    {
        std::stringstream stream;
        stream
            <<  "data used for send is selfoverlapping with the given combination of displacements and counts!" << std::endl;

        myLogger->createMessage(
                MUST_WARNING_SELFOVERLAPPED,
                pId,
                lId,
                MustWarningMessage,
                stream.str()
                );
    }
    // check for overlaps with pending requests
    return checkOverlapsRequests(pId, lId, iList, true, &must::OverlapChecks::outputSendOverlapsRequests);
}

//=============================
// recvOverlapcheckCounts
//=============================
GTI_ANALYSIS_RETURN OverlapChecks::recvOverlapcheckCounts (
        MustParallelId pId,
        MustLocationId lId,
        MustAddressType buffer,
        int displs[],
        int counts[],
        MustDatatypeType datatype,
        int commsize
        )
{
    MustMemIntervalListType iList, tList;
    bool isSelfoverlaping = false;
    I_Datatype * typeinfo = myDatMod->getDatatype(pId, datatype);
    if (typeinfo == NULL)
        return GTI_ANALYSIS_SUCCESS;
    MustAddressType extent = typeinfo->getExtent();
    // get all blocks, used for communication
    for (int i=0; i<commsize; i++)
    {
        tList = calcIntervalList(typeinfo, buffer+displs[i]*extent, counts[i], (MustRequestType)0, false, &isSelfoverlaping);
        iList.insert(tList.begin(), tList.end());
    }

    MustMemIntervalListType::iterator iter, nextIter;

    // selfoverlaping -> error for recv, quit check
    if (isOverlapped(iList, iter, nextIter, true))
    {
        std::stringstream stream;
        stream
            <<  "data used for receive is selfoverlapping with the given combination of displacements and counts!" << std::endl;

        myLogger->createMessage(
                MUST_ERROR_SELFOVERLAPPED,
                pId,
                lId,
                MustErrorMessage,
                stream.str()
                );
        return GTI_ANALYSIS_FAILURE;
    }
    // check for overlaps with pending requests
    return checkOverlapsRequests(pId, lId, iList, false, &must::OverlapChecks::outputRecvOverlapsRequests);
}


//=============================
// sendOverlapcheckTypes
//=============================
GTI_ANALYSIS_RETURN OverlapChecks::sendOverlapcheckTypes (
        MustParallelId pId,
        MustLocationId lId,
        MustAddressType buffer,
        int displs[],
        int counts[],
        MustDatatypeType datatypes[],
        int commsize
        )
{
    MustMemIntervalListType iList, tList;
    I_Datatype * typeinfo;
    bool isSelfoverlaping = false;
    // get all blocks, used for communication
    for (int i=0; i<commsize; i++)
    {
        typeinfo = myDatMod->getDatatype(pId, datatypes[i]);
        if (typeinfo == NULL)
            return GTI_ANALYSIS_SUCCESS;
        tList = calcIntervalList(typeinfo, buffer+displs[i], counts[i], (MustRequestType)0, true, &isSelfoverlaping);
        iList.insert(tList.begin(),tList.end());
    }

    MustMemIntervalListType::iterator iter, nextIter;

    // selfoverlaping -> warning for send, finish check
    if (isOverlapped(iList, iter, nextIter, false))
    {
        std::stringstream stream;
        stream
            <<  "data used for send is selfoverlapping with the given combination of displacements and counts!" << std::endl;

        myLogger->createMessage(
                MUST_WARNING_SELFOVERLAPPED,
                pId,
                lId,
                MustWarningMessage,
                stream.str()
                );
    }
    // check for overlaps with pending requests
    return checkOverlapsRequests(pId, lId, iList, true, &must::OverlapChecks::outputSendOverlapsRequests);
}


//=============================
// recvOverlapcheckTypes
//=============================
GTI_ANALYSIS_RETURN OverlapChecks::recvOverlapcheckTypes (
        MustParallelId pId,
        MustLocationId lId,
        MustAddressType buffer,
        int displs[],
        int counts[],
        MustDatatypeType datatypes[],
        int commsize
        )
{
    MustMemIntervalListType iList, tList;
    bool isSelfoverlaping = false;
    I_Datatype * typeinfo;
    // get all blocks, used for communication
    for (int i=0; i<commsize; i++)
    {
        typeinfo = myDatMod->getDatatype(pId, datatypes[i]);
        if (typeinfo == NULL)
            return GTI_ANALYSIS_SUCCESS;
        tList = calcIntervalList(typeinfo, buffer+displs[i], counts[i], (MustRequestType)0, false, &isSelfoverlaping);
        iList.insert(tList.begin(),tList.end());
    }

    MustMemIntervalListType::iterator iter, nextIter;

    // selfoverlaping -> error for recv, quit check
    if (isOverlapped(iList, iter, nextIter, true))
    {
        std::stringstream stream;
        stream
            <<  "data used for receive is selfoverlapping with the given combination of displacements and counts!" << std::endl;

        myLogger->createMessage(
                MUST_ERROR_SELFOVERLAPPED,
                pId,
                lId,
                MustErrorMessage,
                stream.str()
                );
        return GTI_ANALYSIS_FAILURE;
    }
    // check for overlaps with pending requests
    return checkOverlapsRequests(pId, lId, iList, false, &must::OverlapChecks::outputRecvOverlapsRequests);
}


//=============================
// calculate Intervallist from blocklist
//=============================

MustMemIntervalListType OverlapChecks::calcIntervalList(
        I_Datatype * typeinfo,
        MustAddressType buffer,
        int count,
        MustRequestType request,
        bool isSend,
        bool* hasoverlap)
{
    static MustMemIntervalListType ret;
    static bool myoverlap=false;
    static I_Datatype * lastinfo=NULL;
    static MustAddressType lastbuffer=0;
    static int lastcount=0;
    static MustRequestType lastrequest=0;
    if(typeinfo == lastinfo && lastcount == count && lastbuffer == buffer && request == lastrequest){
        if (hasoverlap != NULL)
            *hasoverlap = *hasoverlap || myoverlap;
        return ret;
    }
    ret.clear();
    lastinfo = typeinfo;
    lastcount = count;
    lastbuffer = buffer;
    { // get the blocklist for the datatype and repeate as requested (count)
        BlockInfo& blockInfo = typeinfo->getBlockInfo();
        myoverlap = blockInfo.overlapped;
        MustAddressType extent = typeinfo->getExtent();
        MustAddressType size = typeinfo->getSize();
        ret = buildMemIntervallist(blockInfo.stridedBlocklist, extent, size, buffer, request, isSend, typeinfo, count);
        if (!myoverlap)
        {
            MustMemIntervalListType::iterator iter, nextIter;
            myoverlap = isOverlapped(ret, iter, nextIter, false);
        }
    }
    if (hasoverlap != NULL)
        *hasoverlap = *hasoverlap || myoverlap;
    return ret;
}

//=============================
// mark prepared blocklist as active
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::makeBlocksActive(
        MustParallelId pId,
        MustLocationId lId,
        MustMemIntervalListType & preparedList,
        MustRequestType request
        )
{
    int rank = pId2Rank(pId);
    requestLocation[rank][request] = std::make_pair(pId, lId);
    MustMemIntervalListType::iterator iter;
    std::list< MustMemIntervalListType::iterator > & activeRequestsBlocklist = activeRequestsBlocklists[rank][request];
    MustMemIntervalListType & myActiveBlocks = activeBlocks[rank];
    MustMemIntervalListType::iterator activIter = myActiveBlocks.begin();
    for (iter = preparedList.begin(); iter != preparedList.end(); iter++)
    {
        // insert into activeBlocks and save the iterator for deletion
        activIter = myActiveBlocks.insert(activIter, *iter);
        activeRequestsBlocklist.push_back(activIter);
    }
    // activeBlocks is always expected to be sorted
//     activeBlocks[rank].sort();
    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// remove blocklist from active list
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::makeBlocksInActive(
        MustParallelId pId,
        MustRequestType request
        )
{
    int rank = pId2Rank(pId);
    return makeBlocksInActive(rank, request);
}

GTI_ANALYSIS_RETURN OverlapChecks::makeBlocksInActive(
        int rank,
        MustRequestType request
        )
{
    // forget the location of a request
    requestLocation[rank].erase(request);
    std::list< MustMemIntervalListType::iterator >::iterator iter;
    for (iter = activeRequestsBlocklists[rank][request].begin(); iter != activeRequestsBlocklists[rank][request].end(); iter++)
    {
        activeBlocks[rank].erase(*iter);
    }
    // activeBlocks is always expected to be sorted: deletion does not touch the sorting

    // after deleting the blocks from activeBlocks, delete the list of bookmarks
    activeRequestsBlocklists[rank].erase(request);
    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// mark prepared blocklist as active
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::checkOverlapsRequests (
        MustParallelId pId,
        MustLocationId lId,
        MustMemIntervalListType & iList,
        bool isSend,
        void (must::OverlapChecks::*outputFunction)(MustParallelId,MustLocationId,MustRequestType))
{
    int rank = pId2Rank(pId), posA, posB;
    MustMemIntervalListType::iterator iterA, iterB;
    if(isOverlapped(activeBlocks[rank], iList, iterA, iterB, posA, posB, true))
    {
                (this->*outputFunction)(pId, lId, iterA->request);
                return GTI_ANALYSIS_FAILURE;
    }
    return GTI_ANALYSIS_SUCCESS;
}

GTI_ANALYSIS_RETURN OverlapChecks::checkOverlapsRequests (
        MustParallelId pId,
        MustLocationId lId,
        MustMemIntervalListType & iList,
        bool isSend,
        const char* errstring, enum MustMessageIdNames errcode)
{
    int rank = pId2Rank(pId), posA, posB;
    MustMemIntervalListType::iterator iterA, iterB;
    if(isOverlapped(activeBlocks[rank], iList, iterA, iterB, posA, posB, true))
    {
        std::list<std::pair<MustParallelId,MustLocationId> > refs;
        std::stringstream stream;

        stream
            <<  errstring << std::endl
            << std::endl;

        I_Request* rInfo = myReqMod->getRequest (pId, iterA->request);
        if (rInfo && !rInfo->isNull())
        {
            stream
                << "(Information on the request associated with the other communication:" << std::endl;
                rInfo->printInfo (stream, &refs);
                stream << ")" << std::endl;
        }

        stream 
            << "(Information on the datatype associated with the other communication:" << std::endl;
        iterA->type->printInfo (stream, &refs);
        stream << ")" << std::endl
            << "The other communication overlaps with this communication at position:";
        iterA->type->printDatatypeLongPos(stream, posA);

        stream
                << std::endl << std::endl;

        rInfo = myReqMod->getRequest (pId, iterB->request);
        if (rInfo && !rInfo->isNull())
        {
            stream
                    << "(Information on the request associated with this communication:" << std::endl;
            rInfo->printInfo (stream, &refs);
            stream << ")" << std::endl;
        }

        stream 
            << "(Information on the datatype associated with this communication:" << std::endl;
        iterB->type->printInfo (stream, &refs);
        stream << ")" << std::endl
            << "This communication overlaps with the other communication at position:";
        iterB->type->printDatatypeLongPos(stream, posB);
        stream  << std::endl;

        myLogger->createMessage(
                errcode,
                pId,
                lId,
                MustErrorMessage,
                stream.str(),
                refs
                );
        return GTI_ANALYSIS_FAILURE;
    }
    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// prepared error message, given to the check as callback
//=============================
void OverlapChecks::outputSendOverlapsRequests(
        MustParallelId pId,
        MustLocationId lId,
        MustRequestType request)
{
    std::list<std::pair<MustParallelId,MustLocationId> > refs;
    std::stringstream stream;
    stream
        <<  "The memory regions to be transfered by this send operation overlap with regions spanned by a pending non-blocking receive operation!" << std::endl
        << "(Information on the request associated with the receive:" << std::endl;
    I_Request* rInfo = myReqMod->getRequest (pId, request);
    if (rInfo) rInfo->printInfo (stream, &refs);
    else stream << "unknown request";
    stream << ")" << std::endl;

    myLogger->createMessage(
            MUST_ERROR_OVERLAPPED_SEND,
            pId,
            lId,
            MustErrorMessage,
            stream.str(),
            refs
            );
}

//=============================
// sendOverlapsRequests
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::sendOverlapsRequests (
        MustParallelId pId,
        MustLocationId lId,
        MustDatatypeType datatype,
        MustAddressType buffer,
        int count)
{
    I_Datatype * typeinfo = myDatMod->getDatatype(pId, datatype);
    if (typeinfo == NULL)
    {
//        std::cout << "Datatype " << datatype << " for rank " << pId2Rank(pId) << " not found in sendOverlapsRequests" << std::endl;
        return GTI_ANALYSIS_SUCCESS;
    }
//    std::cout << "Datatype " << datatype << " for rank " << pId2Rank(pId) << " successfully found in sendOverlapsRequests" << std::endl;
    if (activeBlocks[pId2Rank(pId)].empty())
        return GTI_ANALYSIS_SUCCESS;
    MustMemIntervalListType iList = calcIntervalList(typeinfo, buffer, count, (MustRequestType)0, true);
//     return checkOverlapsRequests(pId, lId, iList, true, &must::OverlapChecks::outputSendOverlapsRequests);
    return checkOverlapsRequests(pId, lId, iList, true, "The memory regions to be transfered by this send operation overlap with regions spanned by a pending non-blocking receive operation!", MUST_ERROR_OVERLAPPED_SEND);
}

//=============================
// prepared error message, given to the check as callback
//=============================

void OverlapChecks::outputRecvOverlapsRequests(
        MustParallelId pId,
        MustLocationId lId,
        MustRequestType request)
{
    std::list<std::pair<MustParallelId,MustLocationId> > refs;
    std::stringstream stream;
    stream
        <<  "The memory regions to be transfered by this receive operation overlap with regions spanned by a pending non-blocking operation!" << std::endl
        << "(Information on the request associated with the other communication:" << std::endl;
    I_Request* rInfo = myReqMod->getRequest (pId, request);
    if (rInfo) rInfo->printInfo (stream, &refs);
    else stream << "unknown request";
    stream << ")" << std::endl;

    myLogger->createMessage(
            MUST_ERROR_OVERLAPPED_RECV,
            pId,
            lId,
            MustErrorMessage,
            stream.str(),
            refs
            );
}

//=============================
// recvOverlapsRequests
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::recvOverlapsRequests (
        MustParallelId pId,
        MustLocationId lId,
        MustDatatypeType datatype,
        MustAddressType buffer,
        int count)
{
    I_Datatype * typeinfo = myDatMod->getDatatype(pId, datatype);
    if (typeinfo == NULL)
        return GTI_ANALYSIS_SUCCESS;
    if (activeBlocks[pId2Rank(pId)].empty())
        return GTI_ANALYSIS_SUCCESS;
    MustMemIntervalListType iList = calcIntervalList(typeinfo, buffer, count, (MustRequestType)0, false);
//     return checkOverlapsRequests(pId, lId, iList, false, &must::OverlapChecks::outputRecvOverlapsRequests);
    return checkOverlapsRequests(pId, lId, iList, false, "The memory regions to be transfered by this receive operation overlap with regions spanned by a pending non-blocking operation!", MUST_ERROR_OVERLAPPED_RECV);
}

//=============================
// overlapsRequests
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::overlapsRequests (
        MustParallelId pId,
        MustLocationId lId,
        MustDatatypeType datatype,
        MustAddressType buffer,
        int count,
        char isSend)
{
    if (isSend)
        return sendOverlapsRequests(pId, lId, datatype, buffer, count);
    else
        return recvOverlapsRequests(pId, lId, datatype, buffer, count);
}

//=============================
// announceSendRequest
//=============================
GTI_ANALYSIS_RETURN OverlapChecks::announceSendRequest (
        MustParallelId pId,
        MustLocationId lId,
        MustDatatypeType datatype,
        MustAddressType buffer,
        int count,
        MustRequestType request)
{
    return announceRequest(pId, lId, datatype, buffer, count, true, request);
}

//=============================
// announceRecvRequest
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::announceRecvRequest (
        MustParallelId pId,
        MustLocationId lId,
        MustDatatypeType datatype,
        MustAddressType buffer,
        int count,
        MustRequestType request)
{
    return announceRequest(pId, lId, datatype, buffer, count, false, request);
}

//=============================
// announceRequest
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::announceRequest (
        MustParallelId pId,
        MustLocationId lId,
        MustDatatypeType datatype,
        MustAddressType buffer,
        int count,
        char isSend,
        MustRequestType request)
{
    I_Datatype * typeinfo = myDatMod->getDatatype(pId, datatype);
    if (typeinfo == NULL)
        return GTI_ANALYSIS_SUCCESS;
    MustMemIntervalListType iList = calcIntervalList(typeinfo, buffer, count, request, isSend!=0);
    makeBlocksActive(pId, lId, iList, request);
    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// announcePSendRequest
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::announcePSendRequest (
        MustParallelId pId,
        MustLocationId lId,
        MustDatatypeType datatype,
        MustAddressType buffer,
        int count,
        MustRequestType request)
{
    return announcePRequest(pId, lId, datatype, buffer, count, true, request);
}

//=============================
// announcePRecvRequest
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::announcePRecvRequest (
        MustParallelId pId,
        MustLocationId lId,
        MustDatatypeType datatype,
        MustAddressType buffer,
        int count,
        MustRequestType request)
{
    return announcePRequest(pId, lId, datatype, buffer, count, false, request);
}

//=============================
// announcePRequest
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::announcePRequest (
        MustParallelId pId,
        MustLocationId lId,
        MustDatatypeType datatype,
        MustAddressType buffer,
        int count,
        char isSend,
        MustRequestType request)
{
    I_Datatype * typeinfo = myDatMod->getDatatype(pId, datatype);
    if (typeinfo == NULL)
        return GTI_ANALYSIS_SUCCESS;
    preparedBlocklists[pId][request] = calcIntervalList(typeinfo, buffer, count, request, isSend!=0);
    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// prepared error message, given to the check as callback
//=============================

void OverlapChecks::outputStartPRequest(
        MustParallelId pId,
        MustLocationId lId,
        MustRequestType request)
{
    std::list<std::pair<MustParallelId,MustLocationId> > refs;
    std::stringstream stream;
    stream
        <<  "A start of a persistent request will transfer memory regions that are still used by a pending non-blocking communication!" << std::endl
        << "(Information on the request that is associated with the later communication:" << std::endl;
    I_Request* rInfo = myReqMod->getRequest (pId, request);
    if (rInfo) rInfo->printInfo (stream, &refs);
    else stream << "unknown request";
    stream << ")" << std::endl;

    myLogger->createMessage(
            MUST_ERROR_OVERLAPPED_SEND,
            pId,
            lId,
            MustErrorMessage,
            stream.str(),
            refs
            );
}

//=============================
// startPRequest
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::startPRequest (
        MustParallelId pId,
        MustLocationId lId,
        MustRequestType request)
{
//     GTI_ANALYSIS_RETURN ret = checkOverlapsRequests(pId, lId, preparedBlocklists[pId][request], preparedBlocklists[pId][request].begin()->isSend, &must::OverlapChecks::outputStartPRequest);
    GTI_ANALYSIS_RETURN ret = checkOverlapsRequests(pId, lId, preparedBlocklists[pId][request], preparedBlocklists[pId][request].begin()->isSend, "A start of a persistent request will transfer memory regions that are still used by a pending non-blocking communication!", MUST_ERROR_OVERLAPPED_SEND);
    makeBlocksActive(pId, lId, preparedBlocklists[pId][request], request);
    return ret;
}

//=============================
// startPRequestArray
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::startPRequestArray (
        MustParallelId pId,
        MustLocationId lId,
        MustRequestType* requests,
        int count)
{
    int i;
    for (i=0; i<count; i++)
        startPRequest(pId, lId, requests[i]);
    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// finishRequest
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::finishRequest (
        MustParallelId pId,
        MustLocationId lId,
        MustRequestType request)
{
    makeBlocksInActive(pId, request);
    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// finishRequests
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::finishRequests (
        MustParallelId pId,
        MustLocationId lId,
        MustRequestType* request,
        int count)
{
    int i;
    for(i=0; i<count; i++)
        makeBlocksInActive(pId, request[i]);
    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// freeRequest
//=============================

GTI_ANALYSIS_RETURN OverlapChecks::freeRequest (
        MustParallelId pId,
        MustLocationId lId,
        MustRequestType request)
{
    static bool firstRun=true;
    I_Request* rInfo;
    if ((firstRun) && ((rInfo = myReqMod->getRequest (pId, request)) != NULL) && (! rInfo->isPersistent()))
    {
        std::list<std::pair<MustParallelId,MustLocationId> > refs;
        std::stringstream stream;
        stream
            <<  "A non-persistent request handle is freed before end of transmission! There will be no further overlap checks for these requests." << std::endl
            << "(Information on the request that is associated with the later communication:" << std::endl;
        rInfo->printInfo (stream, &refs);
        stream << ")" << std::endl;

        myLogger->createMessage(
                MUST_INFO_FREE_NONPERSISTENT_REQUEST,
                pId,
                lId,
                MustInformationMessage,
                stream.str(),
                refs
                );
        firstRun=false;
    }
//     if (preparedBlocklists[pId].find(request) != preparedBlocklists[pId].end())
//     std::cout << "Removing request " << request << " from preparedBlocklists" << std::endl;
    makeBlocksInActive(pId, request);
    preparedBlocklists[pId].erase(request);
    return GTI_ANALYSIS_SUCCESS;
}

