/*
 * MUST (Marmot Umpire Scalable Tool)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2011-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2011-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file DatatypeTrack.h
 *       @see MUST::DatatypeTrack.
 *
 *  @date 10.02.2011
 *  @author Tobias Hilbrich, Joachim Protze
 */

#include "ModuleBase.h"
#include "I_DatatypeTrack.h"
#include "I_BaseConstants.h"
#include "TrackBase.h"
#include "HandleInfoBase.h"
#include "Datatype.h"

#include <map>
#include <vector>
#include <string.h>

#include "DatatypeTrackDerivedStorage.h"

#ifndef DATATYPETRACK_H
#define DATATYPETRACK_H


using namespace gti;

namespace must
{
    
    /**
     * Implementation for I_DatatypeTrack.
     */
    class DatatypeTrack : public TrackBase<Datatype, MustDatatypeType, MustMpiDatatypePredefined, DatatypeTrack, I_DatatypeTrack>
    {
    public:
        /**
         * Constructor.
         * @param instanceName name of this module instance.
         */
        DatatypeTrack (const char* instanceName);

        /**
         * Destructor.
         */
        virtual ~DatatypeTrack (void);

        /**
         * @see I_DatatypeTrack::commit
         */
        GTI_ANALYSIS_RETURN commit (
                MustParallelId pId,
                MustLocationId lId,
                MustDatatypeType type
        );

        /**
         * @see I_DatatypeTrack::free
         */
        GTI_ANALYSIS_RETURN free (
                MustParallelId pId,
                MustLocationId lId,
                MustDatatypeType type
        );

        /**
         * @see I_DatatypeTrack::typeContiguous
         */
        GTI_ANALYSIS_RETURN typeContiguous (
                MustParallelId pId,
                MustLocationId lId,
                int count,
                MustDatatypeType oldType,
                MustDatatypeType newType
        );

        /**
         * @see I_DatatypeTrack::typeHindexed
         */
        GTI_ANALYSIS_RETURN typeHindexed (
                MustParallelId pId,
                MustLocationId lId,
                int count,
                int* arrayOfBlocklengths,
                MustAddressType* arrayOfDisplacements,
                MustDatatypeType oldType,
                MustDatatypeType newType
        );

        /**
         * @see I_DatatypeTrack::typeIndexed
         */
        GTI_ANALYSIS_RETURN typeIndexed (
                MustParallelId pId,
                MustLocationId lId,
                int count,
                int* arrayOfBlocklengths,
                int* arrayOfDisplacements,
                MustDatatypeType oldType,
                MustDatatypeType newType
        );

        /**
         * @see I_DatatypeTrack::typeHvector
         */
        GTI_ANALYSIS_RETURN typeHvector (
                MustParallelId pId,
                MustLocationId lId,
                int count,
                int blocklength,
                MustAddressType stride,
                MustDatatypeType oldType,
                MustDatatypeType newType
        );

        /**
         * @see I_DatatypeTrack::typeVector
         */
        GTI_ANALYSIS_RETURN typeVector (
                MustParallelId pId,
                MustLocationId lId,
                int count,
                int blocklength,
                int stride,
                MustDatatypeType oldType,
                MustDatatypeType newType
        );

        /**
         * @see I_DatatypeTrack::typeStruct
         */
        GTI_ANALYSIS_RETURN typeStruct (
                MustParallelId pId,
                MustLocationId lId,
                int count,
                int* arrayOfBlocklengths,
                MustAddressType* arrayOfDisplacements,
                MustDatatypeType* oldTypes,
                MustDatatypeType newType
        );

        /**
         * @see I_DatatypeTrack::typeIndexedBlock
         */
        GTI_ANALYSIS_RETURN typeIndexedBlock (
                MustParallelId pId,
                MustLocationId lId,
                int count,
                int blocklength,
                int* arrayOfDisplacements,
                MustDatatypeType oldType,
                MustDatatypeType newType
        );

        /**
         * @see I_DatatypeTrack::typeResized
         */
        GTI_ANALYSIS_RETURN typeResized (
                MustParallelId pId,
                MustLocationId lId,
                MustAddressType lb,
                MustAddressType extent,
                MustDatatypeType oldType,
                MustDatatypeType newType
        );

        /**
         * @see I_DatatypeTrack::typeSubarray
         */
        GTI_ANALYSIS_RETURN typeSubarray (
                MustParallelId pId,
                MustLocationId lId,
                int ndims,
                int* arrayOfSizes,
                int* arrayOfSubsizes,
                int* arrayOfStarts,
                int order,
                MustDatatypeType oldType,
                MustDatatypeType newType
        );

        /**
         * @see I_DatatypeTrack::typeDarray
         */
        GTI_ANALYSIS_RETURN typeDarray (
                MustParallelId pId,
                MustLocationId lId,
                int size,
                int rank,
                int ndims,
                int* arrayOfGsizes,
                int* arrayOfDistribs,
                int* arrayOfDargs,
                int* arrayOfPsizes,
                int order,
                MustDatatypeType oldType,
                MustDatatypeType newType
        );

        /**
         * @see I_DatatypeTrack::addPredefinedTypes
         * Overloaded from TrackBase, triggers extra
         * calculations, pushes extents and alignments
         * into base type infos.
         */
        GTI_ANALYSIS_RETURN addPredefinedTypes (
                MustDatatypeType datatypeNull,
                int numPredefs,
                int* predefinedIds,
                MustDatatypeType* predefinedValues,
                MustAddressType *extents,
                int *alignments
        );

        /**
         * @see I_DatatypeTrack::getDatatype
         */
        I_Datatype * getDatatype (
                MustParallelId pId,
                MustDatatypeType type);

        /**
         * @see I_DatatypeTrack::getDatatype
         */
        I_Datatype * getDatatype (
                int rank,
                MustDatatypeType type);

        /**
         * @see I_DatatypeTrack::getPersistentDatatype
         */
        I_DatatypePersistent * getPersistentDatatype (
                MustParallelId pId,
                MustDatatypeType type);

        /**
         * @see I_DatatypeTrack::getPersistentDatatype
         */
        I_DatatypePersistent * getPersistentDatatype (
                int rank,
                MustDatatypeType type);


        /**
         * Returns a pointer to the base constants module.
         * Must not be freed.
         * @return pointer or NULL in case of an error.
         */
        I_BaseConstants* getBCoMod(void);
        int CacheHitCount;
        int CacheMissCount;

        /**
         * Queries the tracker for the information structure for
         * MPI_UB.
         * @return information structure, managed by the tracker.
         */
        FullBaseTypeInfo* getUbInfo (void);

        /**
         * Queries the tracker for the information structure for
         * MPI_LB.
         * @return information structure, managed by the tracker.
         */
        FullBaseTypeInfo* getLbInfo (void);
        
    protected:
        I_BaseConstants* myBCoMod; /**< Module used to query base constants.*/

        std::map <MustMpiDatatypePredefined, FullBaseTypeInfo*> myPredefinedInfos;

        FullBaseTypeInfo   *myUbInfo, /**< Pointer to information structure of MPI_UB.*/
                           *myLbInfo; /**< Pointer to information structure of MPI_LB.*/

        /**
         * Overrides MUST::TrackBase::fillPredefinedInfos.
         */
        void fillPredefinedInfos (void);

        /**
         * Used to initialize null and predefined infos.
         * @see TrackBase::createPredefinedInfo.
         * (Implementation of hook)
         */
        Datatype* createPredefinedInfo (int predef, MustDatatypeType handle);

    public:
        /**
         * Returns the name of a predefined datatype.
         * @param enum value of the predefined.
         * @return textual name, e.g. "MPI_INT".
         */
        std::string getPredefinedName (MustMpiDatatypePredefined predefined);

    }; /*class DatatypeTrack */
} /*namespace MUST*/

#endif /*DATATYPETRACK_H*/
