/*
 * MUST (Marmot Umpire Scalable Tool)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2011-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2011-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file GroupTable.cpp
 *       @see MUST::GroupTable.
 *
 *  @date 07.03.2011
 *  @author Tobias Hilbrich, Mathias Korepkat
 */

#include "GroupTable.h"

#include <assert.h>

using namespace must;

//=============================
// Constructor
//=============================
GroupTable::GroupTable (std::vector<int> set, GroupTrack * track)
	: mySet (set),
	  myRefCount (1),
	  myBeginRank (-1),
	  myEndRank (-1),
	  myTrack (track),
	  myWorld2Rank ()

{
	//nothing to do
}

//=============================
// Constructor
//=============================
GroupTable::GroupTable (int beginRank, int endRank, GroupTrack *track)
	: mySet (),
	  myRefCount (1),
	  myBeginRank (beginRank),
	  myEndRank (endRank),
	  myTrack (track),
	  myWorld2Rank ()

{
	assert (endRank >= beginRank);
}

//=============================
// Destructor
//=============================
GroupTable::~GroupTable (void)
{
	mySet.clear();
	myTrack = NULL;
}

//=============================
// translate
//=============================
bool GroupTable::translate (int rank, int *outWorldRank)
{
	//Representation 2:
	if (myBeginRank >= 0)
	{
		if (rank > myEndRank - myBeginRank)
			return false;

		if (outWorldRank)
			*outWorldRank = myBeginRank + rank;
		return true;
	}

	//Representation 1:
	if (rank >= mySet.size())
		return false;

	if (outWorldRank)
		*outWorldRank = mySet[rank];

	return true;
}

//=============================
// containsWorldRank
//=============================
bool GroupTable::containsWorldRank (int worldRank, int *outGroupRank)
{
    //Representation 2:
    if (myBeginRank >= 0)
    {
        if (worldRank > myEndRank || worldRank < myBeginRank)
            return false;

        if (outGroupRank)
            *outGroupRank = worldRank  - myBeginRank;
        return true;
    }

    //Representation 1:
    if (myWorld2Rank.size () == 0)
    {
        for (int i = 0; i < mySet.size(); i++)
        {
            myWorld2Rank.insert (std::make_pair (mySet[i],i));
        }
    }

    std::map<int, int>::iterator pos = myWorld2Rank.find(worldRank);
    if (pos == myWorld2Rank.end())
        return false;

    if (outGroupRank)
        *outGroupRank = pos->second;
    return true;
}

//=============================
// getSize
//=============================
int GroupTable::getSize ()
{
	//Representation 2:
	if (myBeginRank >= 0)
		return myEndRank - myBeginRank + 1;

	//Representation 1:
	return mySet.size();
}

//=============================
// incrementRefCount
//=============================
int GroupTable::incrementRefCount (void)
{
	myRefCount++;

	//std::cout << "Incremented " << (unsigned long) this <<  " to " << myRefCount << std::endl;

	return myRefCount;
}

//=============================
// decrementRefCount
//=============================
int GroupTable::decrementRefCount (void)
{
	myRefCount--;
	int temp = myRefCount;

	//std::cout << "Decremented " << (unsigned long) this <<  " to " << myRefCount << std::endl;

	if (myRefCount <= 0 && myTrack)
		myTrack->deleteGroupTable (this);

	return temp;
}

//=============================
// isEqual
//=============================
bool GroupTable::isEqual (std::vector<int> *set)
{
	//does the size matches ?
	if (set->size() != getSize())
		return false;

	//Loop over all elements in the given set
	for (int i = 0; i < set->size(); i++)
	{
		int otherWorldRank = (*set)[i];

		if (myBeginRank < 0)
		{
			//Representation 1:
			if (mySet[i] != otherWorldRank)
				return false;
		}
		else
		{
			//Representation 2:
			if (myBeginRank + i != otherWorldRank)
				return false;
		}
	}

	return true;
}

//=============================
// isEqual
//=============================
bool GroupTable::isEqual (int beginRank, int endRank)
{
	//Representation 2:
	if (myBeginRank >= 0)
	{
		return (myBeginRank == beginRank && myEndRank == endRank);
	}

	//Representation 1:
	//Does the size matches ?
	if (getSize() != endRank - beginRank + 1)
		return false;

	//Loop over all elements in this set
	for (int i = 0; i < mySet.size(); i++)
	{
		if (mySet[i] != beginRank + i)
			return false;
	}

	return true;
}

//=============================
// isEqual
//=============================
bool GroupTable::isEqual (I_GroupTable* group)
{
	if (myBeginRank < 0)
		return group->isEqual(&mySet);
	return group->isEqual(myBeginRank, myEndRank);
}

//=============================
// getMapping
//=============================
const std::vector<int>& GroupTable::getMapping (void)
{
    //If we are in representation2, we have to switch to representation1 now
    if (myBeginRank >= 0)
    {
        mySet.resize(myEndRank - myBeginRank + 1);
        for (int i = 0; i <= myEndRank - myBeginRank; i++)
        {
            mySet[i] = myBeginRank + i;
        }

        myBeginRank = -1;
        myEndRank = -1;
    }

    return mySet;
}

/*EOF*/
