/*
 * MUST (Marmot Umpire Scalable Tool)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2010-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2010-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file RequestTrack.cpp
 *       @see MUST::RequestTrack.
 *
 *  @date 24.01.2010
 *  @author Tobias Hilbrich
 */

#include "GtiMacros.h"

#include "RequestTrack.h"

#include <sstream>

using namespace must;

mGET_INSTANCE_FUNCTION(RequestTrack);
mFREE_INSTANCE_FUNCTION(RequestTrack);
mPNMPI_REGISTRATIONPOINT_FUNCTION(RequestTrack);

//=============================
// Constructor
//=============================
RequestTrack::RequestTrack (const char* instanceName)
	: TrackBase<Request, MustRequestType, MustMpiRequestPredefined, RequestTrack, I_RequestTrack> (instanceName)
{
    //Get the DatatypeTrack and CommTrack modules
    if (myFurtherMods.size () < 3)
    {
        std::cerr << "Error: the RequestTrack module needs the DatatypeTrack and CommTrack modules as childs, but at least one of them was not available." << std::endl;
        assert (0);
    }

    myDTrack = (I_DatatypeTrack*) myFurtherMods[0];
    myCTrack = (I_CommTrack*) myFurtherMods[1];
    myConsts = (I_BaseConstants*) myFurtherMods[2];
}

//=============================
// Destructor
//=============================
RequestTrack::~RequestTrack ()
{
    //Nothing to do
}

//=============================
// createPersistentSend
//=============================
GTI_ANALYSIS_RETURN RequestTrack::createPersistentSend (
		MustParallelId pId,
		MustLocationId lId,
		int count,
		MustDatatypeType datatype,
		int dest,
		int tag,
		MustCommType comm,
		int sendMode,
		MustRequestType request)
{
	//==1) Is this request already a known request
	if (isAlreadyKnown(pId, request))
		return GTI_ANALYSIS_FAILURE;

	//==2)  Create the full information record
	Request* info = new Request ();
	info->myIsActive = false;
	info->myIsPersistent = true;
	info->myIsSend = true;
	info->myIsCanceled = false;
	info->myIsNull = false;

	//persistent info, see MustPersistentSendInfo, MustPersistentRecvInfo
	info->myCount = count;
	info->myDatatype = myDTrack->getPersistentDatatype(pId, datatype);
	info->myDestSource = dest;
	info->myTag = tag;
	info->myComm = myCTrack->getPersistentComm(pId, comm);
	info->mySendMode = (MustSendMode) sendMode;

	//Usage history, see MustRequestHistory
	info->myCreationPId = pId;
	info->myCreationLId = lId;

	//Mark if MPI_PROC_NULL
	if (myConsts->getProcNull() == info->myDestSource)
	    info->myIsProcNull = true;
	else
	    info->myIsProcNull = false;

	//==3) Add the full information to myUserHandles
	submitUserHandle (pId, request, info);

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// createPersistentRecv
//=============================
GTI_ANALYSIS_RETURN RequestTrack::createPersistentRecv (
		MustParallelId pId,
		MustLocationId lId,
		int count,
		MustDatatypeType datatype,
		int source,
		int tag,
		MustCommType comm,
		MustRequestType request)
{
	//==1) Is this request already a known request
	if (isAlreadyKnown(pId, request))
		return GTI_ANALYSIS_FAILURE;

	//==2)  Create the full information record
	Request* info = new Request ();
	info->myIsActive = false;
	info->myIsPersistent = true;
	info->myIsSend = false;
	info->myIsCanceled = false;
	info->myIsNull = false;

	//persistent info, see MustPersistentSendInfo, MustPersistentRecvInfo
	info->myCount = count;
	info->myDatatype = myDTrack->getPersistentDatatype(pId, datatype);
	info->myDestSource = source;
	info->myTag = tag;
	info->myComm = myCTrack->getPersistentComm(pId, comm);

	//Usage history, see MustRequestHistory
	info->myCreationPId = pId;
	info->myCreationLId = lId;

	//Mark if MPI_PROC_NULL
	if (myConsts->getProcNull() == info->myDestSource)
	    info->myIsProcNull = true;
	else
	    info->myIsProcNull = false;

	//==3) Add the full information to myUserHandles
	submitUserHandle (pId, request, info);

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// cancel
//=============================
GTI_ANALYSIS_RETURN RequestTrack::cancel (
		MustParallelId pId,
		MustLocationId lId,
		MustRequestType request)
{
	Request* info = getHandleInfo (pId, request);
	if (!info || info->isNull()) return GTI_ANALYSIS_SUCCESS;

	//== Possible errors:
	//1) not known (handled above with return)
	//2) is MPI_REQUEST_NULL (handled above with return)
	//3) is not active
	if (!info->isActive())
		return GTI_ANALYSIS_SUCCESS;
	//4) is already canceled (does not hurts for tracking)

	//== Save the cancel
	info->myIsCanceled = true;
	info->myCancelLId = lId;
	info->myCancelPId = pId;

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// addActive
//=============================
GTI_ANALYSIS_RETURN RequestTrack::addActive (
		MustParallelId pId,
		MustLocationId lId,
		int isSend,
		MustRequestType request,
        int destSource)
{
    //==1) Verify that this request is not already in our map
    Request* reqInfo = getHandleInfo (pId, request);

    if (reqInfo)
    {
        //If already known, it is NULL, EMPTY, or a user defined
        //For NULL and EMPTY nothing to do
        //For user handles, inc the MPI ref count
        if (!reqInfo->isNull())
        {
            reqInfo->mpiIncRefCount();
            reqInfo->myIsActive = true;
        }

        return GTI_ANALYSIS_SUCCESS;
    }

	//==2) Create the full info
	Request* info = new Request ();
	info->myIsActive = true;
	info->myIsPersistent = false;
	info->myIsSend = (bool)isSend;
	info->myIsCanceled = false;
	info->myIsNull = false;

	//==2b) Mark if MPI_PROC_NULL
	if (myConsts->getProcNull() == destSource)
	    info->myIsProcNull = true;
	else
	    info->myIsProcNull = false;

	//Usage history, see MustRequestHistory
	info->myActivationPId = pId;
	info->myActivationLId = lId;

	//==3) Add the full information to myUserHandles
	submitUserHandle (pId, request, info);

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// forceFree
//=============================
GTI_ANALYSIS_RETURN RequestTrack::forceFree (
		MustParallelId pId,
		MustLocationId lId,
		MustRequestType request)
{
	//==1) find the request
    Request* info = getHandleInfo (pId, request);
    if (!info || info->isNull()) return GTI_ANALYSIS_SUCCESS;

	//==2) Potential errors:
	//Is not known (already handled above)
	//Is MPI_REQUEST_NULL (already handled above)

	//==3) Remove the request
    removeUserHandle (pId, request);

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// startPersistent
//=============================
GTI_ANALYSIS_RETURN RequestTrack::startPersistent (
		MustParallelId pId,
		MustLocationId lId,
		MustRequestType request)
{
	//==1) Find the request
    Request* info = getHandleInfo (pId, request);
    if (!info || info->isNull()) return GTI_ANALYSIS_SUCCESS;

	//==2) Potential errors:
	//Not known (handled above)
	//Is MPI_REQUEST_NULL (handled above)
	//Not persistent
	if (!info->myIsPersistent)
		return GTI_ANALYSIS_SUCCESS;
	//Still active
	if (info->myIsActive)
		return GTI_ANALYSIS_SUCCESS;
	//Canceled -> and not completed yet! (We reset canceled below)

	//==3) Handle the start
	info->myIsCanceled = false;
	info->myIsActive = true;
	info->myActivationPId = pId;
	info->myActivationLId = lId;

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// startPersistentArray
//=============================
GTI_ANALYSIS_RETURN RequestTrack::startPersistentArray (
		MustParallelId pId,
		MustLocationId lId,
		MustRequestType* requests,
		int count)
{
	//Loop over all requests, and call startPersistent
	for (int i = 0; i < count; i++)
	{
		if (startPersistent (pId, lId, requests[i]) == GTI_ANALYSIS_FAILURE)
			return GTI_ANALYSIS_FAILURE;
	}

	// Important: set isCanceled to false!
	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// complete
//=============================
GTI_ANALYSIS_RETURN RequestTrack::complete (
		MustParallelId pId,
		MustLocationId lId,
		MustRequestType request,
		int flag)
{
	//==1) Evaluate flag
	if (flag != 1)
		return GTI_ANALYSIS_SUCCESS;

	//==2) Find the request
	Request* info = getHandleInfo (pId, request);
	if (!info || info->isNull()) return GTI_ANALYSIS_SUCCESS;

	//==3) Potential errors:
	//Not known, MPI_REQUEST_NULL -> both handled above
	//Not active
	if (!info->myIsActive)
		return GTI_ANALYSIS_SUCCESS;

	//==4) Handle complete
	info->myIsActive = false;
	info->myIsCanceled = false;

	if (!info->myIsPersistent)
	{
	    removeUserHandle (pId, request);
	}

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// completeAny
//=============================
GTI_ANALYSIS_RETURN RequestTrack::completeAny (
			MustParallelId pId,
			MustLocationId lId,
			MustRequestType* requests,
			int count,
			int index,
			int flag)
{
	//==1) Evaluate flag
	if (flag != 1)
		return GTI_ANALYSIS_SUCCESS;

	//==2) Check index
	if (index < 0 || index >= count)
		return GTI_ANALYSIS_SUCCESS;

	//==3) Propagate to regular complete
	return complete (pId, lId, requests[index], flag);
}

//=============================
// completeArray
//=============================
GTI_ANALYSIS_RETURN RequestTrack::completeArray (
		MustParallelId pId,
		MustLocationId lId,
		MustRequestType *requests,
		int count,
		int flag)
{
	if (!flag)
		return GTI_ANALYSIS_SUCCESS;

	//Loop over all requests
	for (int i = 0; i < count; i++)
	{
		if (complete (pId, lId, requests[i], flag) != GTI_ANALYSIS_SUCCESS)
			return GTI_ANALYSIS_SUCCESS;
	}

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// completeSome
//=============================
GTI_ANALYSIS_RETURN RequestTrack::completeSome (
		MustParallelId pId,
		MustLocationId lId,
		MustRequestType *requests,
		int count,
		int *indices,
		int numIndices)
{
	//Loop over all indices
	for (int i = 0; i < numIndices; i++)
	{
		if (indices[i] >= count)
		{
			std::cerr << "Error: an index in completeSome (" << __FILE__<< "@" << __LINE__ << ") is larger outside the array of indices (indices[" << i << "]=" << indices[i] << ", count=" << count << "). This is an error in the MPI or MUST implementation." << std::endl;
			return GTI_ANALYSIS_FAILURE;
		}

		if (complete (pId, lId, requests[indices[i]], true) != GTI_ANALYSIS_SUCCESS)
			return GTI_ANALYSIS_FAILURE;
	}

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// getRequest
//=============================
I_Request* RequestTrack::getRequest (
		MustParallelId pId,
		MustRequestType request)
{
	return getRequest (pId2Rank(pId), request);
}

//=============================
// getRequest
//=============================
I_Request* RequestTrack::getRequest (
		int rank,
		MustRequestType request)
{
	return getHandleInfo (rank, request);
}

//=============================
// getPersistentRequest
//=============================
I_RequestPersistent* RequestTrack::getPersistentRequest (
        MustParallelId pId,
        MustRequestType request)
{
    return getPersistentRequest (pId2Rank(pId), request);
}

//=============================
// getPersistentRequest
//=============================
I_RequestPersistent* RequestTrack::getPersistentRequest (
        int rank,
        MustRequestType request)
{
    Request* ret = getHandleInfo (rank, request);
    if (ret) ret->incRefCount();
    return ret;
}

//=============================
// createPredefinedInfo
//=============================
Request* RequestTrack::createPredefinedInfo (int value, MustRequestType handle)
{
    if (handle == myNullValue)
        return new Request ();
    return NULL;//There should not be any other cases
}

/*EOF*/
