/*
 * MUST (Marmot Umpire Scalable Tool)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2011-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2011-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file TrackBase.hpp
 *       @see MUST::TrackBase.
 *
 *  @date 10.02.2011
 *  @author Tobias Hilbrich
 */

//=============================
// Constructor
//=============================
template <typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE>
TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::TrackBase (const char* instanceName)
: ModuleBase<SUPER, INTERFACE> (instanceName),
  myNullValue (-1),
  myPredefineds (),
  myUserHandles (),
  myLastQuery (),
  myPIdMod (NULL),
  myFurtherMods ()
  {
    HandleInfoBase::subscribeTracker();

	//create sub modules
	std::vector<I_Module*> subModInstances;
	subModInstances = ModuleBase<SUPER, INTERFACE>::createSubModuleInstances ();

	//handle sub modules
	if (subModInstances.size() < 1)
	{
		std::cerr << "ERROR: " << __FILE__ << "@" << __LINE__ << " needs one sub module as parallel id module." << std::endl;
		assert (0);
	}
	myFurtherMods.resize(subModInstances.size()-1);
	for (int i = 1; i < subModInstances.size(); i++)
	{
		myFurtherMods[i-1] = subModInstances[i];
	}

	myPIdMod = (I_ParallelIdAnalysis*) subModInstances[0];

	//initialize
	myLastQuery = myUserHandles.end();
  }

//=============================
// Destructor
//=============================
template <typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE>
TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::~TrackBase (void)
{
	//destroy ParallelId module
	if (myPIdMod)
		ModuleBase<SUPER, INTERFACE>::destroySubModuleInstance ((I_Module*) myPIdMod);
	myPIdMod = NULL;

	//destroy any other modules
	for (int i = 0; i < myFurtherMods.size(); i++)
	{
		if (myFurtherMods[i])
			ModuleBase<SUPER, INTERFACE>::destroySubModuleInstance (myFurtherMods[i]);
		myFurtherMods[i] = NULL;
	}
	myFurtherMods.clear();

	freeHandleMaps ();

	HandleInfoBase::unsubscribeTracker();
}

//=============================
// addPredefineds
//=============================
template <typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE>
GTI_ANALYSIS_RETURN TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::addPredefineds (
		HANDLE_TYPE nullValue,
		int numPredefs,
		int* predefinedIds,
		HANDLE_TYPE* predefinedValues
)
{
	static bool wasAdded = false;

	if (wasAdded)
		return GTI_ANALYSIS_SUCCESS;
	wasAdded = true;

	//store the null value
	myNullValue = nullValue;
    myNullInfo = createPredefinedInfo(0, nullValue);

	//store the other predefined values
	for (int i = 0; i < numPredefs; i++)
	{
        myPredefineds.insert (std::make_pair(predefinedValues[i], createPredefinedInfo (predefinedIds[i], predefinedValues[i])));
	}

	return GTI_ANALYSIS_SUCCESS;
}

//=============================
// findUserHandle
//=============================
template <typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE>
typename TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::HandleMap::iterator
TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::findUserHandle (
        MustParallelId pId, HANDLE_TYPE handle)
{
    return findUserHandle (pId2Rank(pId), handle);
}

//=============================
// findUserHandle
//=============================
template <typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE>
typename TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::HandleMap::iterator
TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::findUserHandle (
        int rank, HANDLE_TYPE handle)
{
    //Look at last query (avoid the search if that was already the right request)
    if (    myLastQuery != myUserHandles.end()  &&
            myLastQuery->first.first == rank &&
            myLastQuery->first.second == handle)
        return myLastQuery;

    myLastQuery = myUserHandles.find (std::make_pair(rank, handle));

    return myLastQuery;
}

//=============================
// getHandleInfo
//=============================
template <typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE>
FULL_INFO* TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::getHandleInfo (MustParallelId pId, HANDLE_TYPE handle)
{
    return getHandleInfo (pId2Rank(pId), handle);
}

//=============================
// getHandleInfo
//=============================
template <typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE>
FULL_INFO* TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::getHandleInfo (int rank, HANDLE_TYPE handle)
{
    FULL_INFO* ret;

    // is it MPI_handle_NULL ?
    if (handle == myNullValue)
    {
        return myNullInfo;
    }

    //Is it a user defined type ?
    if (    myLastQuery == myUserHandles.end()  ||
            myLastQuery->first.first != rank ||
            myLastQuery->first.second != handle)
    {
        myLastQuery = myUserHandles.find (std::make_pair(rank, handle));
    }

    if (myLastQuery == myUserHandles.end())
    {
        //If not a user type-> it must a predefined

        typename myPredefinedInfos::iterator prePos = myPredefineds.find(handle);

        if (prePos == myPredefineds.end())
            return NULL;

        ret = prePos->second;
    }
    else
    {
        ret = myLastQuery->second;
    }

    return ret;
}

//=============================
// isAlreadyKnown
//=============================
template <typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE>
bool TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::isAlreadyKnown (
		MustParallelId pId, HANDLE_TYPE handle)
{
#ifdef MUST_DEBUG
	//Known as non-predefined request ?
	typename HandleMap::iterator pos = myUserHandles.find (std::make_pair(pId2Rank(pId), handle));
	if (pos != myUserHandles.end())
	{
		//This is more of an internal error, we log it directly to std::cerr
		std::cerr << "Error: in " << __FILE__ << __LINE__ << " tried to add a user handle that is still present, implementation error in MPI or MUST." << std::endl;
		return true;
	}

	//Is null
	if (handle == myNullValue)
	{
		//This is more of an internal error, we log it directly to std::cerr
		std::cerr << "Error: in " << __FILE__ << __LINE__ << " tried to add a handle that had the value of the null handle, implementation error in MPI or MUST." << std::endl;
		return true;
	}

	//Is it a predefined value
	typename myPredefinedInfos::iterator prePos = myPredefineds.find (handle);
	if (prePos != myPredefineds.end())
	{
		//This is more of an internal error, we log it directly to std::cerr
		std::cerr << "Error: in " << __FILE__ << __LINE__ << " tried to add a user handle that has the value of a predefined handle, implementation error in MPI or MUST." << std::endl;
		return true;
	}
#endif /*MUST_DEBUG*/

	return false;
}

//=============================
// submitUserHandle
//=============================
template <typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE>
bool TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::submitUserHandle (
        MustParallelId pId, HANDLE_TYPE handle, FULL_INFO* handleInfo)
{
    std::pair<typename HandleMap::iterator, bool> ret = 
                        myUserHandles.insert(
                            std::make_pair(
                                std::make_pair(pId2Rank(pId), handle),
                                handleInfo
                            )
                        );
    // refCount initialized with 1 in constructor
    myLastQuery = ret.first;
    return ret.second;
}

//=============================
// removeUserHandle
//=============================
template <typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE>
bool TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::removeUserHandle (
        MustParallelId pId, HANDLE_TYPE handle)
{
    typename HandleMap::iterator pos = findUserHandle (pId, handle);
    if (pos == myUserHandles.end())
        return false;

    // decrease MPI refCount (Remove from list if last mpi ref count was removed of no info is present at all)
    if (!pos->second || pos->second->mpiErase())
    {
        myUserHandles.erase(pos);
        myLastQuery = myUserHandles.end();
    }

    return true;
}

//=============================
// pId2Rank
//=============================
template <typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE>
int TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::pId2Rank (
		MustParallelId pId)
{
	return myPIdMod->getInfoForId(pId).rank;
}

//=============================
// createPredefinedInfo
//=============================
template <typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE>
FULL_INFO *
TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::createPredefinedInfo (int value, HANDLE_TYPE handle)
{
	//return an invalid info
    return NULL;
}

//=============================
// getUserHandles
//=============================
template <typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE>
std::list<std::pair<int, HANDLE_TYPE> >
TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::getUserHandles (void)
{
	typename std::list<std::pair<int, HANDLE_TYPE> > handles;

	typename HandleMap::iterator i;
	for (i = myUserHandles.begin(); i != myUserHandles.end(); i++)
	{
		handles.push_back(std::make_pair(i->first.first, i->first.second));
	}

	return handles;
}

template <typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE>
void
TrackBase<FULL_INFO, HANDLE_TYPE, PREDEFINED_ENUM, SUPER, INTERFACE>::freeHandleMaps ()
{
    //free data
    /*
     * We iterate over all handles that are still in out predefined/user handle list.
     * We perform an mpi destroy on each remaining entry:
     * - If the destroy returns that
     *    the object still exist, then some module forgot to erase a persistent
     *    info.
     *
     * Important:
     * If we have such forgotten handles, we simply not free them, we must also
     * no call the "user destroy" function of HandleInfoBase! As handles may have
     * dependencies between each other, freeing some handle before another
     * may cause segfaults or assertions. So we just leave them to exist (which
     * only happens in an error case anyways).
     */
    ////User
    typename HandleMap::iterator userIter = myUserHandles.begin();
    for (; userIter != myUserHandles.end(); userIter++)
    {
        if (!userIter->second) continue;

        if(!userIter->second->mpiDestroy())
        {
#ifdef MUST_DEBUG
            std::cout << "Warning: A module forgot to erase a persistent handle (it was also not freed by the application, thats why we saw this ...)." << std::endl;
#endif
        }
    }

    ////Predefined
    typename myPredefinedInfos::iterator preIter = myPredefineds.begin();
    for (; preIter != myPredefineds.end(); preIter++)
    {
        if (!preIter->second) continue;

        if(!preIter->second->mpiDestroy())
        {
#ifdef MUST_DEBUG
           std::cout << "Warning: some module failed to erase a persistent handle, you should breakpoint here and investigate." << std::endl;
#endif
        }
    }

    ////Null
    if (myNullInfo)
        myNullInfo->mpiDestroy();
    myNullInfo = NULL;

    myPredefineds.clear();
    myUserHandles.clear();
}

/*EOF*/
